"""
Record for EndResult as well as parsing, serialization, and parsing.

When doing experiments with `moptipy`, you apply algorithm setups to problem
instances. For each `setup x instance` combination, you may conduct a series
of repetitions (so-called runs) with different random seeds. Each single run
of an algorithm setup on a problem instances can produce a separate log file.
From each log file, we can load a :class:`EndResult` instance, which
represents, well, the end result of the run, i.e., information such as the
best solution quality reached, when it was reached, and the termination
criterion. These end result records then can be the basis for, e.g., computing
summary statistics via :mod:`~moptipy.evaluation.end_statistics` or for
plotting the end result distribution via
:mod:`~moptipy.evaluation.plot_end_results`.
"""
import argparse
from dataclasses import dataclass
from math import inf, isfinite
from typing import Any, Callable, Final, Iterable, cast

from pycommons.io.console import logger
from pycommons.io.csv import (
    CSV_SEPARATOR,
    SCOPE_SEPARATOR,
    csv_column,
    csv_column_or_none,
    csv_read,
    csv_scope,
    csv_str_or_none,
    csv_val_or_none,
    csv_write,
    pycommons_footer_bottom_comments,
)
from pycommons.io.path import Path, file_path, line_writer
from pycommons.strings.string_conv import (
    int_or_none_to_str,
    num_or_none_to_str,
    num_to_str,
    str_to_num,
)
from pycommons.types import (
    check_int_range,
    check_to_int_range,
    type_error,
)

from moptipy.api.logging import (
    FILE_SUFFIX,
    KEY_ALGORITHM,
    KEY_BEST_F,
    KEY_GOAL_F,
    KEY_INSTANCE,
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_MAX_FES,
    KEY_MAX_TIME_MILLIS,
    KEY_RAND_SEED,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
    PROGRESS_CURRENT_F,
    PROGRESS_FES,
    PROGRESS_TIME_MILLIS,
    SECTION_PROGRESS,
)
from moptipy.evaluation._utils import (
    _check_max_time_millis,
)
from moptipy.evaluation.base import (
    DESC_ALGORITHM,
    DESC_ENCODING,
    DESC_INSTANCE,
    DESC_OBJECTIVE_FUNCTION,
    F_NAME_NORMALIZED,
    F_NAME_RAW,
    F_NAME_SCALED,
    KEY_ENCODING,
    KEY_OBJECTIVE_FUNCTION,
    PerRunData,
    motipy_footer_bottom_comments,
)
from moptipy.evaluation.log_parser import SetupAndStateParser
from moptipy.utils.help import moptipy_argparser
from moptipy.utils.math import try_float_div, try_int, try_int_div
from moptipy.utils.strings import (
    sanitize_names,
)

#: a description of the random seed
DESC_RAND_SEED: Final[str] = (
    "the value of the seed of the random number generator used in the run. "
    f"Random seeds are in 0..{int((1 << (8 * 8)) - 1)} and the random "
    f"number generators are those from numpy.")
#: the description of best-F
DESC_BEST_F: Final[str] = (
    " the best (smallest) objective value ever encountered during the run ("
    "regardless whether the algorithm later forgot it again or not).")
#: the description of the last improvement FE
DESC_LAST_IMPROVEMENT_FE: Final[str] = (
    "the objective function evaluation (FE) when the last improving move took"
    " place. 1 FE corresponds to the construction and evaluation "
    "of one solution. The first FE has index 1. With 'last "
    "improving move' we mean the last time when a solution was "
    "discovered that was better than all previous solutions. This "
    "time / FE index is the one when the solution with objective "
    f"value {KEY_BEST_F} was discovered.")
#: the description of the last improvement time milliseconds
DESC_LAST_IMPROVEMENT_TIME_MILLIS: Final[str] = (
    "the clock time in milliseconds after the begin of the run when "
    "the last improving search move took place.")
#: the description of the total FEs
DESC_TOTAL_FES: Final[str] = (
    "the total number of objective function evaluations (FEs) that were "
    "performed during the run.")
#: the total consumed time in milliseconds
DESC_TOTAL_TIME_MILLIS: Final[str] = (
    "the clock time in milliseconds that has passed between the begin of the "
    "run and the end of the run.")
#: the description of the goal objective value
DESC_GOAL_F: Final[str] = (
    "the goal objective value. A run will stop as soon as a solution was"
    "discovered which has an objective value less than or equal to "
    f"{KEY_GOAL_F}. In other words, as soon as {KEY_BEST_F} reaches or dips "
    f"under {KEY_GOAL_F}, the algorithm will stop. If {KEY_GOAL_F} is not "
    "reached, the run will continue until other budget limits are exhausted. "
    "If a lower bound for the objective function is known, this is often used"
    " as a goal objective value. If o goal objective value is specified, this"
    " field is empty.")
#: a description of the budget as the maximum objective function evaluation
DESC_MAX_FES: Final[str] = (
    "the maximum number of permissible FEs per run. As soon as this limit is "
    f"reached, the run will stop. In other words, {KEY_TOTAL_FES} will never "
    f"be more than {KEY_MAX_FES}. A run may stop earlier if some other "
    "termination criterion is reached, but never later.")
#: a description of the budget in terms of maximum runtime
DESC_MAX_TIME_MILLIS: Final[str] = (
    "the maximum number of milliseconds of clock time that a run is permitted"
    " to use as computational budget before being terminated. This limit is "
    "more of a soft limit, as we cannot physically stop a run at arbitrary "
    "points without causing mayhem. Thus, it may be that some runs consume "
    "slightly more runtime than this limit. But the rule is that the "
    "algorithm gets told to stop (via should_terminate() becoming True) as "
    f"soon as this time has elapsed. But generally, {KEY_TOTAL_TIME_MILLIS}<="
    f"{KEY_MAX_TIME_MILLIS} approximately holds.")


@dataclass(frozen=True, init=False, order=False, eq=False)
class EndResult(PerRunData):
    """
    An immutable end result record of one run of one algorithm on one problem.

    This record provides the information of the outcome of one application of
    one algorithm to one problem instance in an immutable way.
    """

    #: The best objective value encountered.
    best_f: int | float

    #: The index of the function evaluation when best_f was reached.
    last_improvement_fe: int

    #: The time when best_f was reached.
    last_improvement_time_millis: int

    #: The total number of performed FEs.
    total_fes: int

    #: The total time consumed by the run.
    total_time_millis: int

    #: The goal objective value if provided
    goal_f: int | float | None

    #: The (optional) maximum permitted FEs.
    max_fes: int | None

    #: The (optional) maximum runtime.
    max_time_millis: int | None

    def __init__(self,
                 algorithm: str,
                 instance: str,
                 objective: str,
                 encoding: str | None,
                 rand_seed: int,
                 best_f: int | float,
                 last_improvement_fe: int,
                 last_improvement_time_millis: int,
                 total_fes: int,
                 total_time_millis: int,
                 goal_f: int | float | None,
                 max_fes: int | None,
                 max_time_millis: int | None):
        """
        Create a consistent instance of :class:`EndResult`.

        :param algorithm: the algorithm name
        :param instance: the instance name
        :param objective: the name of the objective function
        :param encoding: the name of the encoding that was used, if any, or
            `None` if no encoding was used
        :param rand_seed: the random seed
        :param best_f: the best reached objective value
        :param last_improvement_fe: the FE when best_f was reached
        :param last_improvement_time_millis: the time when best_f was reached
        :param total_fes: the total FEs
        :param total_time_millis: the total runtime
        :param goal_f: the goal objective value, if provide
        :param max_fes: the optional maximum FEs
        :param max_time_millis: the optional maximum runtime

        :raises TypeError: if any parameter has a wrong type
        :raises ValueError: if the parameter values are inconsistent
        """
        super().__init__(algorithm, instance, objective, encoding, rand_seed)
        object.__setattr__(self, "best_f", try_int(best_f))
        object.__setattr__(
            self, "last_improvement_fe", check_int_range(
                last_improvement_fe, "last_improvement_fe",
                1, 1_000_000_000_000_000))
        object.__setattr__(
            self, "last_improvement_time_millis", check_int_range(
                last_improvement_time_millis, "last_improvement_time_millis",
                0, 100_000_000_000))
        object.__setattr__(
            self, "total_fes", check_int_range(
                total_fes, "total_fes", last_improvement_fe,
                1_000_000_000_000_000))
        object.__setattr__(
            self, "total_time_millis", check_int_range(
                total_time_millis, "total_time_millis",
                last_improvement_time_millis, 100_000_000_000))

        if goal_f is not None:
            goal_f = None if goal_f <= -inf else try_int(goal_f)
        object.__setattr__(self, "goal_f", goal_f)

        if max_fes is not None:
            check_int_range(max_fes, "max_fes", total_fes,
                            1_000_000_000_000_000_000)
        object.__setattr__(self, "max_fes", max_fes)

        if max_time_millis is not None:
            check_int_range(
                max_time_millis, "max_time_millis", 1, 100_000_000_000)
            _check_max_time_millis(max_time_millis,
                                   total_fes,
                                   total_time_millis)
        object.__setattr__(self, "max_time_millis", max_time_millis)

    def _tuple(self) -> tuple[Any, ...]:
        """
        Get the tuple representation of this object used in comparisons.

        :return: the comparison-relevant data of this object in a tuple
        """
        return (self.__class__.__name__,
                "" if self.algorithm is None else self.algorithm,
                "" if self.instance is None else self.instance,
                "" if self.objective is None else self.objective,
                "" if self.encoding is None else self.encoding,
                1, self.rand_seed, "", "",
                inf if self.goal_f is None else self.goal_f,
                inf if self.max_fes is None else self.max_fes,
                inf if self.max_time_millis is None else self.max_time_millis,
                self.best_f, self.last_improvement_fe,
                self.last_improvement_time_millis, self.total_fes,
                self.total_time_millis)

    def success(self) -> bool:
        """
        Check if a run is successful.

        This method returns `True` if and only if `goal_f` is defined and
        `best_f <= goal_f` (and `False` otherwise).

        :return: `True` if and only if `best_f<=goal_f`
        """
        return False if self.goal_f is None else self.best_f <= self.goal_f

    def path_to_file(self, base_dir: str) -> Path:
        """
        Get the path that would correspond to the log file of this end result.

        Obtain a path that would correspond to the log file of this end
        result, resolved from a base directory `base_dir`.

        :param base_dir: the base directory
        :returns: the path to a file corresponding to the end result record
        """
        return Path(base_dir).resolve_inside(
            self.algorithm).resolve_inside(self.instance).resolve_inside(
            sanitize_names([self.algorithm, self.instance,
                            hex(self.rand_seed)]) + FILE_SUFFIX)

    def get_best_f(self) -> int | float:
        """
        Get the best objective value reached.

        :returns: the best objective value reached
        """
        if not isinstance(self, EndResult):
            raise type_error(self, "self", EndResult)
        return self.best_f

    def get_last_improvement_fe(self) -> int:
        """
        Get the index of the function evaluation when `best_f` was reached.

        :returns: the index of the function evaluation when `best_f` was
            reached
        """
        if not isinstance(self, EndResult):
            raise type_error(self, "self", EndResult)
        return self.last_improvement_fe

    def get_last_improvement_time_millis(self) -> int:
        """
        Get the milliseconds when `best_f` was reached.

        :returns: the milliseconds when `best_f` was reached
        """
        if not isinstance(self, EndResult):
            raise type_error(self, "self", EndResult)
        return self.last_improvement_time_millis

    def get_total_fes(self) -> int:
        """
        Get the total number of performed FEs.

        :returns: the total number of performed FEs
        """
        if not isinstance(self, EndResult):
            raise type_error(self, "self", EndResult)
        return self.total_fes

    def get_total_time_millis(self) -> int:
        """
        Get the total time consumed by the run.

        :returns: the total time consumed by the run
        """
        if not isinstance(self, EndResult):
            raise type_error(self, "self", EndResult)
        return self.total_time_millis

    def get_goal_f(self) -> int | float | None:
        """
        Get the goal objective value, if any.

        :returns: the goal objective value, if any
        """
        if not isinstance(self, EndResult):
            raise type_error(self, "self", EndResult)
        return self.goal_f

    def get_max_fes(self) -> int | None:
        """
        Get the maximum number of FEs permissible.

        :returns: the maximum number of FEs permissible
        """
        if not isinstance(self, EndResult):
            raise type_error(self, "self", EndResult)
        return self.max_fes

    def get_max_time_millis(self) -> int | None:
        """
        Get the maximum permissible milliseconds permitted.

        :returns: the maximum permissible milliseconds permitted
        """
        if not isinstance(self, EndResult):
            raise type_error(self, "self", EndResult)
        return self.max_time_millis

    def get_normalized_best_f(self) -> int | float | None:
        """
        Get the normalized f.

        :returns: the normalized f
        """
        g: Final[int | float | None] = EndResult.get_goal_f(self)
        if (g is None) or (g <= 0):
            return None
        return try_float_div(self.best_f - g, g)

    def get_scaled_best_f(self) -> int | float | None:
        """
        Get the normalized f.

        :returns: the normalized f
        """
        g: Final[int | float | None] = EndResult.get_goal_f(self)
        if (g is None) or (g <= 0):
            return None
        return try_float_div(self.best_f, g)

    def get_fes_per_time_milli(self) -> int | float:
        """
        Get the fes per time milliseconds.

        :returns: the fes per time milliseconds
        """
        return try_int_div(EndResult.get_total_fes(self), max(
            1, EndResult.get_total_time_millis(self)))


#: A set of getters for accessing variables of the end result
__PROPERTIES: Final[Callable[[str], Callable[[
    EndResult], int | float | None]]] = {
    KEY_LAST_IMPROVEMENT_FE: EndResult.get_last_improvement_fe,
    "last improvement FE": EndResult.get_last_improvement_fe,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS:
        EndResult.get_last_improvement_time_millis,
    "last improvement ms": EndResult.get_last_improvement_time_millis,
    KEY_TOTAL_FES: EndResult.get_total_fes,
    "fes": EndResult.get_total_fes,
    KEY_TOTAL_TIME_MILLIS: EndResult.get_total_time_millis,
    "ms": EndResult.get_total_time_millis,
    KEY_GOAL_F: EndResult.get_goal_f,
    F_NAME_RAW: EndResult.get_best_f,
    KEY_BEST_F: EndResult.get_best_f,
    "f": EndResult.get_best_f,
    F_NAME_SCALED: EndResult.get_scaled_best_f,
    "bestFscaled": EndResult.get_scaled_best_f,
    F_NAME_NORMALIZED: EndResult.get_normalized_best_f,
    "bestFnormalized": EndResult.get_normalized_best_f,
    KEY_MAX_FES: EndResult.get_max_fes,
    "budgetFEs": EndResult.get_max_fes,
    KEY_MAX_TIME_MILLIS: EndResult.get_max_time_millis,
    "budgetMS": EndResult.get_max_time_millis,
    "fesPerTimeMilli": EndResult.get_fes_per_time_milli,
}.get


def getter(dimension: str) -> Callable[[EndResult], int | float | None]:
    """
    Produce a function that obtains the given dimension from EndResults.

    The following dimensions are supported:

    1. `lastImprovementFE`: :attr:`~EndResult.last_improvement_fe`
    2. `lastImprovementTimeMillis`:
        :attr:`~EndResult.last_improvement_time_millis`
    3. `totalFEs`: :attr:`~EndResult.total_fes`
    4. `totalTimeMillis`: :attr:`~EndResult.total_time_millis`
    5. `goalF`: :attr:`~EndResult.goal_f`
    6.  `plainF`, `bestF`: :attr:`~EndResult.best_f`
    7. `scaledF`: :attr:`~EndResult.best_f`/:attr:`~EndResult.goal_f`
    8. `normalizedF`: (:attr:`~EndResult.best_f`-attr:`~EndResult.goal_f`)/
        :attr:`~EndResult.goal_f`
    9. `maxFEs`: :attr:`~EndResult.max_fes`
    10. `maxTimeMillis`: :attr:`~EndResult.max_time_millis`
    11. `fesPerTimeMilli`:  :attr:`~EndResult.total_fes`
        /:attr:`~EndResult.total_time_millis`

    :param dimension: the dimension
    :returns: a callable that returns the value corresponding to the
        dimension from its input value, which must be an :class:`EndResult`
    """
    result: Callable[[EndResult], int | float] | None = __PROPERTIES(
        str.strip(dimension))
    if result is None:
        raise ValueError(f"Unknown EndResult dimension {dimension!r}.")
    return result


def from_logs(
        path: str, consumer: Callable[[EndResult], Any],
        max_fes: int | None | Callable[
            [str, str], int | None] = None,
        max_time_millis: int | None | Callable[
            [str, str], int | None] = None,
        goal_f: int | float | None | Callable[
            [str, str], int | float | None] = None) -> None:
    """
    Parse a given path and pass all end results found to the consumer.

    If `path` identifies a file with suffix `.txt`, then this file is
    parsed. The appropriate :class:`EndResult` is created and appended to
    the `collector`. If `path` identifies a directory, then this directory
    is parsed recursively for each log file found, one record is passed to
    the `consumer`. As `consumer`, you could pass any `callable` that
    accepts instances of :class:`EndResult`, e.g., the `append` method of
    a :class:`list`.

    Via the parameters `max_fes`, `max_time_millis`, and `goal_f`, you can
    set virtual limits for the objective function evaluations, the maximum
    runtime, and the objective value. The :class:`EndResult` records will
    then not represent the actual final state of the runs but be
    synthesized from the logged progress information. This, of course,
    requires such information to be present. It will also raise a
    `ValueError` if the goals are invalid, e.g., if a runtime limit is
    specified that is before the first logged points.

    There is one caveat when specifying `max_time_millis`: Let's say that
    the log files only log improvements. Then you might have a log point
    for 7000 FEs, 1000ms, and f=100. The next log point could be 8000 FEs,
    1200ms, and f=90. Now if your time limit specified is 1100ms, we know
    that the end result is f=100 (because f=90 was reached too late) and
    that the total runtime is 1100ms, as this is the limit you specified
    and it was also reached. But we do not know the number of consumed
    FEs. We know you consumed at least 7000 FEs, but you did not consume
    8000 FEs. It would be wrong to claim that 7000 FEs were consumed,
    since it could have been more. We therefore set a virtual end point at
    7999 FEs. In terms of performance metrics such as the
    :mod:`~moptipy.evaluation.ert`, this would be the most conservative
    choice in that it does not over-estimate the speed of the algorithm.
    It can, however, lead to very big deviations from the actual values.
    For example, if your algorithm quickly converged to a local optimum
    and there simply is no log point that exceeds the virtual time limit
    but the original run had a huge FE-based budget while your virtual
    time limit was small, this could lead to an estimate of millions of
    FEs taking part within seconds...

    :param path: the path to parse
    :param consumer: the consumer
    :param max_fes: the maximum FEs, a callable to compute the maximum
        FEs from the algorithm and instance name, or `None` if unspecified
    :param max_time_millis: the maximum runtime in milliseconds, a
        callable to compute the maximum runtime from the algorithm and
        instance name, or `None` if unspecified
    :param goal_f: the goal objective value, a callable to compute the
        goal objective value from the algorithm and instance name, or
        `None` if unspecified
    """
    need_goals: bool = False
    if max_fes is not None:
        if not callable(max_fes):
            max_fes = check_int_range(
                max_fes, "max_fes", 1, 1_000_000_000_000_000)
        need_goals = True
    if max_time_millis is not None:
        if not callable(max_time_millis):
            max_time_millis = check_int_range(
                max_time_millis, "max_time_millis", 1, 1_000_000_000_000)
        need_goals = True
    if goal_f is not None:
        if callable(goal_f):
            need_goals = True
        else:
            if not isinstance(goal_f, int | float):
                raise type_error(goal_f, "goal_f", (int, float, None))
            if isfinite(goal_f):
                need_goals = True
            elif goal_f <= -inf:
                goal_f = None
            else:
                raise ValueError(f"goal_f={goal_f} is not permissible.")
    if need_goals:
        __InnerProgressLogParser(
            max_fes, max_time_millis, goal_f, consumer).parse(path)
    else:
        __InnerLogParser(consumer).parse(path)


def to_csv(results: Iterable[EndResult], file: str) -> Path:
    """
    Write a sequence of end results to a file in CSV format.

    :param results: the end results
    :param file: the path
    :return: the path of the file that was written
    """
    path: Final[Path] = Path(file)
    logger(f"Writing end results to CSV file {path!r}.")
    path.ensure_parent_dir_exists()
    with path.open_for_write() as wt:
        csv_write(
            data=sorted(results), consumer=line_writer(wt),
            setup=CsvWriter().setup,
            get_column_titles=CsvWriter.get_column_titles,
            get_row=CsvWriter.get_row,
            get_header_comments=CsvWriter.get_header_comments,
            get_footer_comments=CsvWriter.get_footer_comments,
            get_footer_bottom_comments=CsvWriter.get_footer_bottom_comments)
    logger(f"Done writing end results to CSV file {path!r}.")
    return path


def from_csv(file: str, consumer: Callable[[EndResult], Any],
             filterer: Callable[[EndResult], bool]
             = lambda x: True) -> None:
    """
    Parse a given CSV file to get :class:`EndResult` Records.

    :param file: the path to parse
    :param consumer: the collector, can be the `append` method of a
        :class:`list`
    :param filterer: an optional filter function
    """
    if not callable(consumer):
        raise type_error(consumer, "consumer", call=True)
    path: Final[Path] = file_path(file)
    logger(f"Now reading CSV file {path!r}.")

    def __cons(r: EndResult, __c=consumer, __f=filterer) -> None:
        """Consume a record."""
        if __f(r):
            __c(r)

    with path.open_for_read() as rd:
        csv_read(rows=rd,
                 setup=CsvReader,
                 parse_row=CsvReader.parse_row,
                 consumer=__cons)

    logger(f"Done reading CSV file {path!r}.")


class CsvWriter:
    """A class for CSV writing of :class:`EndResult`."""

    def __init__(self, scope: str | None = None) -> None:
        """
        Initialize the csv writer.

        :param scope: the prefix to be pre-pended to all columns
        """
        #: an optional scope
        self.scope: Final[str | None] = (
            str.strip(scope)) if scope is not None else None

        #: has this writer been set up?
        self.__setup: bool = False
        #: do we need the encoding?
        self.__needs_encoding: bool = False
        #: do we need the max FEs?
        self.__needs_max_fes: bool = False
        #: do we need the max millis?
        self.__needs_max_ms: bool = False
        #: do we need the goal F?
        self.__needs_goal_f: bool = False

    def setup(self, data: Iterable[EndResult]) -> "CsvWriter":
        """
        Set up this csv writer based on existing data.

        :param data: the data to setup with
        :returns: this writer
        """
        if self.__setup:
            raise ValueError(
                "EndResults CsvWriter has already been set up.")
        self.__setup = True

        no_encoding: bool = True
        no_max_fes: bool = True
        no_max_ms: bool = True
        no_goal_f: bool = True
        check: int = 4
        for er in data:
            if no_encoding and (er.encoding is not None):
                no_encoding = False
                self.__needs_encoding = True
                check -= 1
                if check <= 0:
                    return self
            if no_max_fes and (er.max_fes is not None):
                self.__needs_max_fes = True
                no_max_fes = False
                check -= 1
                if check <= 0:
                    return self
            if no_max_ms and (er.max_time_millis is not None):
                self.__needs_max_ms = True
                no_max_ms = False
                check -= 1
                if check <= 0:
                    return self
            if no_goal_f and (er.goal_f is not None) and (
                    isfinite(er.goal_f)):
                self.__needs_goal_f = True
                no_goal_f = False
                check -= 1
                if check <= 0:
                    return self
        return self

    def get_column_titles(self, dest: Callable[[str], None]) -> None:
        """
        Get the column titles.

        :param dest: the destination string consumer
        """
        p: Final[str] = self.scope
        dest(csv_scope(p, KEY_ALGORITHM))
        dest(csv_scope(p, KEY_INSTANCE))
        dest(csv_scope(p, KEY_OBJECTIVE_FUNCTION))
        if self.__needs_encoding:
            dest(csv_scope(p, KEY_ENCODING))
        dest(csv_scope(p, KEY_RAND_SEED))
        dest(csv_scope(p, KEY_BEST_F))
        dest(csv_scope(p, KEY_LAST_IMPROVEMENT_FE))
        dest(csv_scope(p, KEY_LAST_IMPROVEMENT_TIME_MILLIS))
        dest(csv_scope(p, KEY_TOTAL_FES))
        dest(csv_scope(p, KEY_TOTAL_TIME_MILLIS))

        if self.__needs_goal_f:
            dest(csv_scope(p, KEY_GOAL_F))
        if self.__needs_max_fes:
            dest(csv_scope(p, KEY_MAX_FES))
        if self.__needs_max_ms:
            dest(csv_scope(p, KEY_MAX_TIME_MILLIS))

    def get_row(self, data: EndResult,
                dest: Callable[[str], None]) -> None:
        """
        Render a single end result record to a CSV row.

        :param data: the end result record
        :param dest: the string consumer
        """
        dest(data.algorithm)
        dest(data.instance)
        dest(data.objective)
        if self.__needs_encoding:
            dest(data.encoding if data.encoding else "")
        dest(hex(data.rand_seed))
        dest(num_to_str(data.best_f))
        dest(str(data.last_improvement_fe))
        dest(str(data.last_improvement_time_millis))
        dest(str(data.total_fes))
        dest(str(data.total_time_millis))
        if self.__needs_goal_f:
            dest(num_or_none_to_str(data.goal_f))
        if self.__needs_max_fes:
            dest(int_or_none_to_str(data.max_fes))
        if self.__needs_max_ms:
            dest(int_or_none_to_str(data.max_time_millis))

    def get_header_comments(self, dest: Callable[[str], None]) -> None:
        """
        Get any possible header comments.

        :param dest: the destination
        """
        dest("Experiment End Results")
        dest("See the description at the bottom of the file.")

    def get_footer_comments(self, dest: Callable[[str], None]) -> None:
        """
        Get any possible footer comments.

        :param dest: the destination
        """
        dest("")
        scope: Final[str | None] = self.scope
        dest("Records describing the end results of single runs ("
             "single executions) of algorithms applied to optimization "
             "problems.")
        dest("Each run is characterized by an algorithm setup, a problem "
             "instance, and a random seed.")
        if scope:
            dest("All end result records start with prefix "
                 f"{scope}{SCOPE_SEPARATOR}.")
        dest(f"{csv_scope(scope, KEY_ALGORITHM)}: {DESC_ALGORITHM}")
        dest(f"{csv_scope(scope, KEY_INSTANCE)}: {DESC_INSTANCE}")
        dest(f"{csv_scope(scope, KEY_OBJECTIVE_FUNCTION)}:"
             f" {DESC_OBJECTIVE_FUNCTION}")
        if self.__needs_encoding:
            dest(f"{csv_scope(scope, KEY_ENCODING)}: {DESC_ENCODING}")
        dest(f"{csv_scope(scope, KEY_RAND_SEED)}: {DESC_RAND_SEED}")
        dest(f"{csv_scope(scope, KEY_BEST_F)}: {DESC_BEST_F}")
        dest(f"{csv_scope(scope, KEY_LAST_IMPROVEMENT_FE)}: "
             f"{DESC_LAST_IMPROVEMENT_FE}")
        dest(f"{csv_scope(scope, KEY_LAST_IMPROVEMENT_TIME_MILLIS)}: "
             f"{DESC_LAST_IMPROVEMENT_TIME_MILLIS}")
        dest(f"{csv_scope(scope, KEY_TOTAL_FES)}: {DESC_TOTAL_FES}")
        dest(f"{csv_scope(scope, KEY_TOTAL_TIME_MILLIS)}: "
             f"{DESC_TOTAL_TIME_MILLIS}")
        if self.__needs_goal_f:
            dest(f"{csv_scope(scope, KEY_GOAL_F)}: {DESC_GOAL_F}")
        if self.__needs_max_fes:
            dest(f"{csv_scope(scope, KEY_MAX_FES)}: {DESC_MAX_FES}")
        if self.__needs_max_ms:
            dest(f"{csv_scope(scope, KEY_MAX_TIME_MILLIS)}: "
                 f"{DESC_MAX_TIME_MILLIS}")

    def get_footer_bottom_comments(self, dest: Callable[[str], None]) -> None:
        """
        Get the footer bottom comments.

        :param dest: the destination to write to.
        """
        motipy_footer_bottom_comments(
            self, dest, ("The end results data is produced using module "
                         "moptipy.evaluation.end_results."))
        pycommons_footer_bottom_comments(self, dest)


class CsvReader:
    """A csv parser for end results."""

    def __init__(self, columns: dict[str, int]) -> None:
        """
        Create a CSV parser for :class:`EndResult`.

        :param columns: the columns
        """
        super().__init__()
        if not isinstance(columns, dict):
            raise type_error(columns, "columns", dict)
        #: the index of the algorithm column, if any
        self.__idx_algorithm: Final[int] = csv_column(columns, KEY_ALGORITHM)
        #: the index of the instance column, if any
        self.__idx_instance: Final[int] = csv_column(columns, KEY_INSTANCE)
        #: the index of the objective function column, if any
        self.__idx_objective: Final[int] = csv_column(
            columns, KEY_OBJECTIVE_FUNCTION)
        #: the index of the encoding column, if any
        self.__idx_encoding = csv_column_or_none(columns, KEY_ENCODING)

        #: the index of the random seed column
        self.__idx_seed: Final[int] = csv_column(columns, KEY_RAND_SEED)
        #: the column with the last improvement FE
        self.__idx_li_fe: Final[int] = csv_column(
            columns, KEY_LAST_IMPROVEMENT_FE)
        #: the column with the last improvement time milliseconds
        self.__idx_li_ms: Final[int] = csv_column(
            columns, KEY_LAST_IMPROVEMENT_TIME_MILLIS)
        #: the column with the best obtained objective value
        self.__idx_best_f: Final[int] = csv_column(columns, KEY_BEST_F)
        #: the column with the total time in FEs
        self.__idx_tt_fe: Final[int] = csv_column(columns, KEY_TOTAL_FES)
        #: the column with the total time in milliseconds
        self.__idx_tt_ms: Final[int] = csv_column(
            columns, KEY_TOTAL_TIME_MILLIS)

        #: the column with the goal objective value, if any
        self.__idx_goal_f: Final[int | None] = csv_column_or_none(
            columns, KEY_GOAL_F)
        #: the column with the maximum FEs, if any such budget constraint was
        #: defined
        self.__idx_max_fes: Final[int | None] = csv_column_or_none(
            columns, KEY_MAX_FES)
        #: the column with the maximum runtime in milliseconds, if any such
        #: budget constraint was defined
        self.__idx_max_ms: Final[int | None] = csv_column_or_none(
            columns, KEY_MAX_TIME_MILLIS)

    def parse_row(self, data: list[str]) -> EndResult:
        """
        Parse a row of data.

        :param data: the data row
        :return: the end result statistics
        """
        return EndResult(
            data[self.__idx_algorithm],  # algorithm
            data[self.__idx_instance],  # instance
            data[self.__idx_objective],  # objective
            csv_str_or_none(data, self.__idx_encoding),  # encoding
            int((data[self.__idx_seed])[2:], 16),  # rand seed
            str_to_num(data[self.__idx_best_f]),  # best_f
            int(data[self.__idx_li_fe]),  # last_improvement_fe
            int(data[self.__idx_li_ms]),  # last_improvement_time_millis
            int(data[self.__idx_tt_fe]),  # total_fes
            int(data[self.__idx_tt_ms]),  # total_time_millis
            csv_val_or_none(data, self.__idx_goal_f, str_to_num),
            csv_val_or_none(data, self.__idx_max_fes, int),  # max_fes
            csv_val_or_none(data, self.__idx_max_ms, int))  # max_time_ms


class __InnerLogParser(SetupAndStateParser):
    """The internal log parser class."""

    def __init__(self, consumer: Callable[[EndResult], Any]):
        """
        Create the internal log parser.

        :param consumer: the consumer accepting the parsed data
        """
        super().__init__()
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        self.__consumer: Final[Callable[[EndResult], Any]] = consumer

    def process(self) -> None:
        self.__consumer(EndResult(self.algorithm,
                                  self.instance,
                                  self.objective,
                                  self.encoding,
                                  self.rand_seed,
                                  self.best_f,
                                  self.last_improvement_fe,
                                  self.last_improvement_time_millis,
                                  self.total_fes,
                                  self.total_time_millis,
                                  self.goal_f,
                                  self.max_fes,
                                  self.max_time_millis))


def _join_goals(vlimit, vgoal, select):  # noqa
    if vlimit is None:
        return vgoal
    if vgoal is None:
        return vlimit
    return select(vlimit, vgoal)


class __InnerProgressLogParser(SetupAndStateParser):
    """The internal log parser class for virtual end results."""

    def __init__(
            self,
            max_fes: int | None | Callable[[str, str], int | None],
            max_time_millis: int | None | Callable[[str, str], int | None],
            goal_f: int | float | None | Callable[
                [str, str], int | float | None],
            consumer: Callable[[EndResult], Any]):
        """
        Create the internal log parser.

        :param consumer: the consumer
        :param max_fes: the maximum FEs, or `None` if unspecified
        :param max_time_millis: the maximum runtime in milliseconds, or
            `None` if unspecified
        :param goal_f: the goal objective value, or `None` if unspecified
        """
        super().__init__()
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        self.__consumer: Final[Callable[[EndResult], Any]] = consumer

        self.__src_limit_ms: Final[
            int | None | Callable[[str, str], int | None]] = max_time_millis
        self.__src_limit_fes: Final[
            int | None | Callable[[str, str], int | None]] = max_fes
        self.__src_limit_f: Final[
            int | float | None | Callable[
                [str, str], int | float | None]] = goal_f

        self.__limit_ms: int | float = inf
        self.__limit_ms_n: int | None = None
        self.__limit_fes: int | float = inf
        self.__limit_fes_n: int | None = None
        self.__limit_f: int | float = -inf
        self.__limit_f_n: int | float | None = None

        self.__stop_fes: int | None = None
        self.__stop_ms: int | None = None
        self.__stop_f: int | float | None = None
        self.__stop_li_fe: int | None = None
        self.__stop_li_ms: int | None = None
        self.__hit_goal: bool = False
        self.__state: int = 0

    def end_file(self) -> bool:
        if self.__state != 2:
            raise ValueError(
                "Illegal state, log file must have a "
                f"{SECTION_PROGRESS!r} section.")
        self.__state = 0
        return super().end_file()

    def process(self) -> None:
        hit_goal = self.__hit_goal
        stop_fes: int = self.__stop_fes
        stop_ms: int = self.__stop_ms
        if not hit_goal:
            stop_ms = max(stop_ms, cast(int, min(
                self.total_time_millis, self.__limit_ms)))
            ul_fes = self.total_fes
            if stop_ms < self.total_time_millis:
                ul_fes = ul_fes - 1
            stop_fes = max(stop_fes, cast(int, min(
                ul_fes, self.__limit_fes)))

        self.__consumer(EndResult(
            algorithm=self.algorithm,
            instance=self.instance,
            objective=self.objective,
            encoding=self.encoding,
            rand_seed=self.rand_seed,
            best_f=self.__stop_f,
            last_improvement_fe=self.__stop_li_fe,
            last_improvement_time_millis=self.__stop_li_ms,
            total_fes=stop_fes,
            total_time_millis=stop_ms,
            goal_f=_join_goals(self.__limit_f_n, self.goal_f, max),
            max_fes=_join_goals(self.__limit_fes_n, self.max_fes, min),
            max_time_millis=_join_goals(
                self.__limit_ms_n, self.max_time_millis, min)))
        self.__stop_fes = None
        self.__stop_ms = None
        self.__stop_f = None
        self.__stop_li_fe = None
        self.__stop_li_ms = None
        self.__limit_fes_n = None
        self.__limit_fes = inf
        self.__limit_ms_n = None
        self.__limit_ms = inf
        self.__limit_f_n = None
        self.__limit_f = -inf
        self.__hit_goal = False

    def start_file(self, path: Path) -> bool:
        if super().start_file(path):
            if (self.algorithm is None) or (self.instance is None):
                raise ValueError(
                    f"Invalid state: algorithm={self.algorithm!r}, "
                    f"instance={self.instance!r}.")

            fes = self.__src_limit_fes(self.algorithm, self.instance) \
                if callable(self.__src_limit_fes) else self.__src_limit_fes
            self.__limit_fes_n = None if fes is None else \
                check_int_range(fes, "limit_fes", 1, 1_000_000_000_000_000)
            self.__limit_fes = inf if self.__limit_fes_n is None \
                else self.__limit_fes_n

            time = self.__src_limit_ms(self.algorithm, self.instance) \
                if callable(self.__src_limit_ms) else self.__src_limit_ms
            self.__limit_ms_n = None if time is None else \
                check_int_range(time, "limit_ms", 1, 1_000_000_000_000)
            self.__limit_ms = inf if self.__limit_ms_n is None \
                else self.__limit_ms_n

            self.__limit_f_n = self.__src_limit_f(
                self.algorithm, self.instance) \
                if callable(self.__src_limit_f) else self.__src_limit_f
            if self.__limit_f_n is not None:
                if not isinstance(self.__limit_f_n, int | float):
                    raise type_error(self.__limit_f_n, "limit_f", (
                        int, float))
                if not isfinite(self.__limit_f_n):
                    if self.__limit_f_n <= -inf:
                        self.__limit_f_n = None
                    else:
                        raise ValueError(
                            f"invalid limit f={self.__limit_f_n} for "
                            f"{self.algorithm} on {self.instance}")
            self.__limit_f = -inf if self.__limit_f_n is None \
                else self.__limit_f_n
            return True
        return False

    def start_section(self, title: str) -> bool:
        if title == SECTION_PROGRESS:
            if self.__state != 0:
                raise ValueError(f"Already did section {title}.")
            self.__state = 1
            return True
        return super().start_section(title)

    def needs_more_lines(self) -> bool:
        return (self.__state < 2) or super().needs_more_lines()

    def lines(self, lines: list[str]) -> bool:
        if not isinstance(lines, list):
            raise type_error(lines, "lines", list)

        if self.__state != 1:
            return super().lines(lines)
        self.__state = 2

        n_rows = len(lines)
        if n_rows < 2:
            raise ValueError("lines must contain at least two elements,"
                             f"but contains {n_rows}.")

        columns = [c.strip() for c in lines[0].split(CSV_SEPARATOR)]
        fe_col: Final[int] = columns.index(PROGRESS_FES)
        ms_col: Final[int] = columns.index(PROGRESS_TIME_MILLIS)
        f_col: Final[int] = columns.index(PROGRESS_CURRENT_F)
        current_fes: int = -1
        current_ms: int = -1
        current_f: int | float = inf
        current_li_fe: int | None = None
        current_li_ms: int | None = None
        stop_fes: int | None = None
        stop_ms: int | None = None
        stop_f: int | float | None = None
        stop_li_fe: int | None = None
        stop_li_ms: int | None = None
        limit_fes: Final[int | float] = self.__limit_fes
        limit_ms: Final[int | float] = self.__limit_ms
        limit_f: Final[int | float] = self.__limit_f

        for line in lines[1:]:
            values = line.split(CSV_SEPARATOR)
            current_fes = check_to_int_range(
                values[fe_col], "fes", current_fes, 1_000_000_000_000_000)
            current_ms = check_to_int_range(
                values[ms_col], "ms", current_ms, 1_000_000_000_00)
            f: int | float = str_to_num(values[f_col])
            if (current_fes <= limit_fes) and (current_ms <= limit_ms):
                if f < current_f:  # can only update best within budget
                    current_f = f
                    current_li_fe = current_fes
                    current_li_ms = current_ms
                stop_ms = current_ms
                stop_fes = current_fes
                stop_f = current_f
                stop_li_fe = current_li_fe
                stop_li_ms = current_li_ms
            if (current_fes >= limit_fes) or (current_ms >= limit_ms) or \
                    (current_f <= limit_f):
                self.__hit_goal = True
                break  # we can stop parsing the stuff

        if (stop_fes is None) or (stop_ms is None) or (stop_f is None) \
                or (current_fes <= 0) or (not isfinite(current_f)):
            raise ValueError(
                "Illegal state, no fitting data point found: stop_fes="
                f"{stop_fes}, stop_ms={stop_ms}, stop_f={stop_f}, "
                f"current_fes={current_fes}, current_ms={current_ms}, "
                f"current_f={current_f}.")

        if current_fes >= limit_fes:
            stop_fes = max(stop_fes, min(
                cast(int, limit_fes), current_fes))
        elif current_ms > limit_ms:
            stop_fes = max(stop_fes, current_fes - 1)
        else:
            stop_fes = max(stop_fes, current_fes)

        if current_ms >= limit_ms:
            stop_ms = max(stop_ms, min(cast(int, limit_ms), current_ms))
        else:
            stop_ms = max(stop_ms, current_ms)

        self.__stop_fes = stop_fes
        self.__stop_ms = stop_ms
        self.__stop_f = stop_f
        self.__stop_li_fe = stop_li_fe
        self.__stop_li_ms = stop_li_ms
        return self.needs_more_lines()


# Run log files to end results if executed as script
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipy_argparser(
        __file__,
        "Convert log files obtained with moptipy to the end results CSV "
        "format that can be post-processed or exported to other tools.",
        "This program recursively parses a folder hierarchy created by"
        " the moptipy experiment execution facility. This folder "
        "structure follows the scheme of algorithm/instance/log_file "
        "and has one log file per run. As result of the parsing, one "
        "CSV file (where columns are separated by ';') is created with"
        " one row per log file. This row contains the end-of-run state"
        " loaded from the log file. Whereas the log files may store "
        "the complete progress of one run of one algorithm on one "
        "problem instance as well as the algorithm configuration "
        "parameters, instance features, system settings, and the final"
        " results, the end results CSV file will only represent the "
        "final result quality, when it was obtained, how long the runs"
        " took, etc. This information is much denser and smaller and "
        "suitable for importing into other tools such as Excel or for "
        "postprocessing.")
    parser.add_argument(
        "source", nargs="?", default="./results",
        help="the location of the experimental results, i.e., the root folder "
             "under which to search for log files", type=Path)
    parser.add_argument(
        "dest", help="the path to the end results CSV file to be created",
        type=Path, nargs="?", default="./evaluation/end_results.txt")
    parser.add_argument(
        "--maxFEs", help="the maximum permitted FEs",
        type=int, nargs="?", default=None)
    parser.add_argument(
        "--maxTime", help="the maximum permitted time in milliseconds",
        type=int, nargs="?", default=None)
    parser.add_argument(
        "--goalF", help="the goal objective value",
        type=str_to_num, nargs="?", default=None)
    args: Final[argparse.Namespace] = parser.parse_args()

    end_results: Final[list[EndResult]] = []
    from_logs(args.source, end_results.append,
              args.maxFEs, args.maxTime, args.goalF)
    to_csv(end_results, args.dest)



"""Compute the end result of one mock run."""

from dataclasses import dataclass
from math import ceil
from typing import Final

from numpy.random import Generator
from pycommons.io.console import logger
from pycommons.types import check_int_range, type_error

from moptipy.evaluation.end_results import EndResult
from moptipy.mock.components import (
    Algorithm,
    BasePerformance,
    Experiment,
    Instance,
)
from moptipy.utils.nputils import rand_generator


def end_result(performance: BasePerformance, seed: int,
               max_fes: int | None = None,
               max_time_millis: int | None = None) -> EndResult:
    """
    Compute the end result of a mock run.

    :param performance: the performance record
    :param seed: the random seed
    :param max_fes: the maximum number of FEs
    :param max_time_millis: the maximum time
    :returns: the end result record
    """
    if not isinstance(performance, BasePerformance):
        raise type_error(performance, "performance", BasePerformance)

    limit_time: int = 1_000_000_000_000
    limit_fes: int = 1_000_000_000_000
    if max_time_millis is not None:
        limit_time = check_int_range(
            max_time_millis, "max_time_millis", 11, 1_000_000_000_000)
    if max_fes is not None:
        limit_fes = check_int_range(
            max_fes, "max_fes", 11, 1_000_000_000_000)

    # The random number generator is determined by the seed.
    random: Final[Generator] = rand_generator(seed)

    # The speed also has some slight jitter.
    jitter: Final[float] = performance.jitter
    speed: float = -1
    while (speed <= 0) or (speed >= 1):
        speed = random.normal(loc=performance.speed, scale=0.01 * jitter)

    # total_time ~ total_fes * (performance.speed ** 3)
    total_time: int
    total_fes: int
    trials: int
    if max_time_millis is not None:
        total_time = int(max_time_millis + abs(random.normal(
            loc=0, scale=5 * jitter)))
        total_fes = -1
        trials = 0
        while ((total_fes <= 100) or (total_fes > limit_fes)) \
                and (trials < 10000):
            trials += 1
            total_fes = int(random.normal(
                loc=max(10.0, total_time / (speed ** 3)),
                scale=max(100.0, 200.0 / speed)))
        if trials >= 10000:
            total_fes = int(min(limit_fes, 10000.0))
    else:
        total_fes = max_fes if max_fes is not None else 1_000_000
        total_time = -1
        trials = 0
        while ((total_time <= 10) or (total_time > limit_time)) \
                and (trials < 10000):
            total_time = int(random.normal(
                loc=max(10.0, total_fes * (speed ** 3)),
                scale=max(10.0, 100.0 / speed)))
        if trials >= 10000:
            total_time = int(min(limit_time, 10000))

    # We now look for the vicinity of the local optimum that will be found.
    # We use the quality to determine which attractor to use.
    # Then we will sample a solution between the next lower and next higher
    # attractor, again using the jitter and quality.

    # First, add some jitter to the quality.
    qual: float = -1
    while (qual <= 0) or (qual >= 1):
        qual = random.normal(loc=performance.performance, scale=0.02 * jitter)

    # Second, find the right attractor and remember it in base.
    att: Final[tuple[int, ...]] = performance.instance.attractors
    attn: Final[int] = len(att)
    att_index: int = -1
    best: Final[int] = performance.instance.best
    worst: Final[int] = performance.instance.worst
    att_trials: int = 1000
    while (att_index < 0) or (att_index >= (attn - 1)):
        att_trials -= 1
        if att_trials <= 0:
            att_index = attn // 2
            break
        att_index = int(random.normal(loc=attn * (qual ** 1.7),
                                      scale=jitter ** 0.9))
    base: Final[int] = att[att_index]

    # Third, choose the ends of the intervals in which we can jitter.
    jit_end: int = min(int(base + 0.6 * (att[att_index + 1] - base)), worst)
    jit_start: int = base
    if att_index > 0:
        jit_start = int(0.5 + ceil(base - 0.6 * (base - att[att_index - 1])))
    jit_start = max(jit_start, best)

    # Now determine best_f.
    best_f: int = -1
    while (best_f < jit_start) or (best_f > jit_end) \
            or (best_f < best) or (best_f > worst):
        uni: float = -1
        while (uni <= 0) or (uni >= 1):
            uni = abs(random.normal(loc=0, scale=jitter))
        best_f = int(round(base - uni * (base - jit_start))) \
            if random.uniform(low=0, high=1) < qual else \
            int(round(base + uni * (jit_end - base)))

    # Finally, we need to compute the time we have used.
    fact: float = -1
    while (fact <= 0) or (fact >= 1):
        fact = 1 - random.exponential(scale=(att_index + 1) / (attn + 1))

    last_improvement_fe: int = -1
    while (last_improvement_fe <= 0) or (last_improvement_fe >= total_fes):
        last_improvement_fe = int(random.normal(
            loc=total_fes * fact, scale=total_fes * 0.05 * jitter))

    last_improvement_time: int = -1
    while (last_improvement_time <= 0) \
            or (last_improvement_time >= total_time):
        last_improvement_time = int(random.normal(
            loc=total_time * fact, scale=total_time * 0.05 * jitter))

    res: Final[EndResult] = EndResult(
        algorithm=performance.algorithm.name,
        instance=performance.instance.name,
        objective="f", encoding="e",
        rand_seed=seed,
        best_f=best_f,
        last_improvement_fe=last_improvement_fe,
        last_improvement_time_millis=last_improvement_time,
        total_fes=total_fes,
        total_time_millis=total_time,
        goal_f=performance.instance.best,
        max_fes=max_fes,
        max_time_millis=max_time_millis)
    return res


@dataclass(frozen=True, init=False, order=True)
class EndResults:
    """An immutable set of end results."""

    #: The experiment.
    experiment: Experiment
    #: The end results.
    results: tuple[EndResult, ...]
    #: The maximum permitted FEs.
    max_fes: int | None
    #: The maximum permitted milliseconds.
    max_time_millis: int | None
    #: the results per algorithm
    __results_for_algo: dict[str | Algorithm, tuple[EndResult, ...]]
    #: the results per instance
    __results_for_inst: dict[str | Instance, tuple[EndResult, ...]]

    def __init__(self,
                 experiment: Experiment,
                 results: tuple[EndResult, ...],
                 max_fes: int | None = None,
                 max_time_millis: int | None = None):
        """
        Create a mock results of an experiment.

        :param experiment: the experiment
        :param results: the end results
        :param max_fes: the maximum permitted FEs
        :param max_time_millis: the maximum permitted milliseconds.
        """
        if not isinstance(experiment, Experiment):
            raise type_error(experiment, "experiment", Experiment)
        object.__setattr__(self, "experiment", experiment)

        per_algo: Final[dict[str | Algorithm, list[EndResult]]] = {}
        per_inst: Final[dict[str | Instance, list[EndResult]]] = {}
        if not isinstance(results, tuple):
            raise type_error(results, "results", tuple)
        if len(results) <= 0:
            raise ValueError("end_results must not be empty.")
        for a in results:
            if not isinstance(a, EndResult):
                raise type_error(a, "element of results", EndResult)
            aa = experiment.get_algorithm(a.algorithm)
            if aa in per_algo:
                per_algo[aa].append(a)
            else:
                per_algo[aa] = [a]
            ii = experiment.get_instance(a.instance)
            if ii in per_inst:
                per_inst[ii].append(a)
            else:
                per_inst[ii] = [a]

        object.__setattr__(self, "results", results)

        pa: dict[str | Algorithm, tuple[EndResult, ...]] = {}
        for ax in experiment.algorithms:
            lax: list[EndResult] = per_algo[ax]
            lax.sort()
            pa[ax.name] = pa[ax] = tuple(lax)
        pi: dict[str | Instance, tuple[EndResult, ...]] = {}
        for ix in experiment.instances:
            lix: list[EndResult] = per_inst[ix]
            lix.sort()
            pi[ix.name] = pi[ix] = tuple(lix)

        object.__setattr__(self, "_EndResults__results_for_algo", pa)
        object.__setattr__(self, "_EndResults__results_for_inst", pi)

        if max_fes is not None:
            check_int_range(max_fes, "max_fes", 1, 1_000_000_000_000)
        object.__setattr__(self, "max_fes", max_fes)

        if max_time_millis is not None:
            check_int_range(
                max_time_millis, "max_time_millis", 1, 1_000_000_000_000_000)
        object.__setattr__(self, "max_time_millis", max_time_millis)

    @staticmethod
    def create(experiment: Experiment,
               max_fes: int | None = None,
               max_time_millis: int | None = None) -> "EndResults":
        """
        Create the end results for a given experiment.

        :param experiment: the experiment
        :param max_fes: the maximum number of FEs
        :param max_time_millis: the maximum time
        :returns: the end results
        """
        if not isinstance(experiment, Experiment):
            raise type_error(experiment, "experiment", Experiment)
        logger(
            "now creating all end results for an experiment with "
            f"{len(experiment.algorithms)} algorithms, "
            f"{len(experiment.instances)} instances, and "
            f"{len(experiment.per_instance_seeds[0])} runs per setup.")

        if max_fes is not None:
            check_int_range(max_fes, "max_fes", 1, 1_000_000_000_000)

        if max_time_millis is not None:
            check_int_range(
                max_time_millis, "max_time_millis", 1, 1_000_000_000_000_000)
        results: list[EndResult] = []
        for per in experiment.applications:
            for seed in experiment.seeds_for_instance(per.instance):
                results.append(end_result(performance=per,
                                          seed=seed,
                                          max_fes=max_fes,
                                          max_time_millis=max_time_millis))
        results.sort()

        res: Final[EndResults] = EndResults(experiment=experiment,
                                            results=tuple(results),
                                            max_fes=max_fes,
                                            max_time_millis=max_time_millis)
        logger(f"finished creating all {len(res.results)} end results.")
        return res

    def results_for_algorithm(self, algorithm: str | Algorithm) \
            -> tuple[EndResult, ...]:
        """
        Get the end results per algorithm.

        :param algorithm: the algorithm
        :returns: the end results
        """
        return self.__results_for_algo[algorithm]

    def results_for_instance(self, instance: str | Instance) \
            -> tuple[EndResult, ...]:
        """
        Get the end results per instance.

        :param instance: the instance
        :returns: the end results
        """
        return self.__results_for_inst[instance]



"""A mock-up of an objective function."""

from math import inf, isfinite, nextafter
from typing import Any, Final, Iterable, cast

import numpy as np
from numpy.random import Generator, default_rng
from pycommons.strings.string_conv import num_to_str
from pycommons.types import type_error

from moptipy.api.objective import Objective
from moptipy.mock.utils import make_ordered_list, sample_from_attractors
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection
from moptipy.utils.nputils import is_np_float, is_np_int


class MockObjective(Objective):
    """A mock-up of an objective function."""

    def __init__(self,
                 is_int: bool = True,
                 lb: int | float = -inf,
                 ub: int | float = inf,
                 fmin: int | float | None = None,
                 fattractors: Iterable[int | float | None] | None = None,
                 fmax: int | float | None = None,
                 seed: int | None = None) -> None:
        """
        Create a mock objective function.

        :param is_int: is this objective function always integer?
        :param lb: the lower bound
        :param ub: the upper bound
        :param fmin: the minimum value this objective actually takes on
        :param fattractors: the attractor points
        :param fmax: the maximum value this objective actually takes on
        """
        if not isinstance(is_int, bool):
            raise type_error(is_int, "is_int", bool)
        #: is this objective integer?
        self.is_int: Final[bool] = is_int

        if seed is None:
            seed = int(default_rng().integers(0, 1 << 63))
        elif not isinstance(seed, int):
            raise type_error(seed, "seed", int)
        #: the random seed
        self.seed: Final[int] = seed

        #: the generator for setting up the mock objective
        random: Final[Generator] = default_rng(seed)
        #: the name of this objective function
        self.name: Final[str] = \
            f"mock{hex(random.integers(1, 100_000_000))[2:]}"

        if not isinstance(lb, int | float):
            raise type_error(lb, "lb", (int, float))
        if isfinite(lb) and is_int and not isinstance(lb, int):
            raise type_error(lb, f"finite lb @ is_int={is_int}", int)
        if not isinstance(ub, int | float):
            raise type_error(ub, "ub", (int, float))
        if isfinite(ub) and is_int and not isinstance(ub, int):
            raise type_error(ub, f"finite lb @ is_int={is_int}", int)
        if lb >= ub:
            raise ValueError(f"lb={lb} >= ub={ub} not permitted")
        #: the lower bound
        self.lb: Final[int | float] = lb
        #: the upper bound
        self.ub: Final[int | float] = ub

        if fmin is not None:
            if not isinstance(fmin, int if is_int else (int, float)):
                raise type_error(fmin, f"fmin[is_int={is_int}",
                                 int if is_int else (int, float))
            if fmin < lb:
                raise ValueError(f"fmin={fmin} < lb={lb}")
        if fmax is not None:
            if not isinstance(fmax, int if is_int else (int, float)):
                raise type_error(fmax, f"fmax[is_int={is_int}",
                                 int if is_int else (int, float))
            if fmax > ub:
                raise ValueError(f"fmax={fmax} < ub={ub}")
        if (fmin is not None) and (fmax is not None) and (fmin >= fmax):
            raise ValueError(f"fmin={fmin} >= fmax={fmax}")

        values: list[int | float | None] = [lb, fmin]
        if fattractors is None:
            while True:
                values.append(None)
                if random.integers(2) <= 0:
                    break
        else:
            values.extend(fattractors)
        values.append(fmax)
        values.append(ub)

        values = make_ordered_list(values, is_int, random)
        if values is None:
            raise ValueError(
                f"could not create mock objective with lb={lb}, fmin={fmin}, "
                f"fattractors={fattractors}, fmax={fmax}, ub={ub}, "
                f"is_int={is_int}, and seed={seed}")

        #: the minimum value the function actually takes on
        self.fmin: Final[int | float] = values[1]
        #: the maximum value the function actually takes on
        self.fmax: Final[int | float] = values[-2]
        #: the mean value the function actually takes on
        self.fattractors: Final[tuple[int | float, ...]] =\
            cast(tuple[int | float, ...], tuple(values[2:-2]))
        #: the internal random number generator
        self.__random: Final[Generator] = random

    def sample(self) -> int | float:
        """
        Sample the mock objective function.

        :returns: the value of the mock objective function
        """
        return sample_from_attractors(self.__random, self.fattractors,
                                      self.is_int, self.lb, self.ub)

    def evaluate(self, x) -> float | int:
        """
        Return a mock objective value.

        :param x: the candidate solution
        :return: the objective value
        """
        seed: int | None = None
        if hasattr(x, "__hash__") and (x.__hash__ is not None):
            seed = hash(x)
        elif isinstance(x, np.ndarray):
            seed = hash(x.tobytes())
        elif isinstance(x, list):
            seed = hash(str(x))
        random = self.__random if seed is None else default_rng(abs(seed))

        return sample_from_attractors(random, self.fattractors,
                                      self.is_int, self.lb, self.ub)

    def lower_bound(self) -> float | int:
        """
        Get the lower bound of the objective value.

        :return: the lower bound of the objective value
        """
        return self.lb

    def upper_bound(self) -> float | int:
        """
        Get the upper bound of the objective value.

        :return: the upper bound of the objective value
        """
        return self.ub

    def is_always_integer(self) -> bool:
        """
        Return `True` if :meth:`~evaluate` will always return an `int` value.

        :returns: `True` if :meth:`~evaluate` will always return an `int`
          or `False` if also a `float` may be returned.
        """
        return self.is_int

    def __str__(self):
        """Get the name of this mock objective function."""
        return self.name

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """Log the special parameters of tis mock objective function."""
        super().log_parameters_to(logger)
        logger.key_value("min", self.fmin)
        logger.key_value("attractors", CSV_SEPARATOR.join([
            num_to_str(n) for n in self.fattractors]))
        logger.key_value("max", self.fmax)
        logger.key_value("seed", self.seed)
        logger.key_value("is_int", self.is_int)

    @staticmethod
    def for_type(dtype: np.dtype) -> "MockObjective":
        """
        Create a mock objective function with values bound by a given `dtype`.

        :param dtype: the numpy data type
        :returns: the mock objective function
        """
        if not isinstance(dtype, np.dtype):
            raise type_error(dtype, "dtype", np.dtype)

        random = default_rng()
        params: dict[str, Any] = {}
        use_min = random.integers(2) <= 0
        use_max = random.integers(2) <= 0
        if not (use_min or use_max):
            if random.integers(5) <= 0:
                use_min = True
            else:
                use_max = True
        if is_np_int(dtype):
            params["is_int"] = True
            ii = np.iinfo(dtype)
            params["lb"] = lbi = max(int(ii.min), -(1 << 58))
            params["ub"] = ubi = min(int(ii.max), (1 << 58))
            if use_min:
                params["fmin"] = lbi + 1
            if use_max:
                params["fmax"] = ubi - 1

        if is_np_float(dtype):
            params["is_int"] = False
            fi = np.finfo(dtype)
            params["lb"] = lbf = max(float(fi.min), -1e300)
            params["ub"] = ubf = min(float(fi.max), 1e300)
            if use_min:
                params["fmin"] = nextafter(float(lbf + fi.eps), inf)
            if use_max:
                params["fmax"] = nextafter(float(ubf - fi.eps), -inf)

        if len(params) > 0:
            return MockObjective(**params)
        raise ValueError(f"unsupported dtype: {dtype}")



"""
A package with example problems and operators.

You can find many more applications and examples at in our `moptipyapps`
package building on `moptipy` at <https://thomasweise.github.io/moptipyapps>.
Here, we mainly include examples that are useful for verifying the algorithms
and modules implemented in `moptipy` and those useful for our "Optimization
Algorithms" book (see <https://thomasweise.github.io/oa>).
"""



"""The multi-objective algorithm execution API."""

from math import isfinite
from typing import Final, TypeVar, cast

from pycommons.types import check_int_range

from moptipy.api._mo_process_no_ss import _MOProcessNoSS
from moptipy.api._mo_process_no_ss_log import _MOProcessNoSSLog
from moptipy.api._mo_process_ss import _MOProcessSS
from moptipy.api._mo_process_ss_log import _MOProcessSSLog
from moptipy.api.algorithm import Algorithm, check_algorithm
from moptipy.api.encoding import Encoding, check_encoding
from moptipy.api.execution import Execution, TExecution
from moptipy.api.mo_archive import MOArchivePruner, check_mo_archive_pruner
from moptipy.api.mo_problem import (
    MOProblem,
    MOSOProblemBridge,
    check_mo_problem,
)
from moptipy.api.mo_process import MOProcess
from moptipy.api.objective import Objective, check_objective
from moptipy.api.process import (
    check_goal_f,
    check_max_fes,
    check_max_time_millis,
)
from moptipy.api.space import Space, check_space
from moptipy.mo.archive.keep_farthest import KeepFarthest
from moptipy.utils.nputils import rand_seed_check

#: The execution type variable for returning `Self`.
TMOExecution = TypeVar("TMOExecution", bound="MOExecution")


class MOExecution(Execution):
    """
    Define all the components of a multi-objective experiment and execute it.

    Different from :class:`~moptipy.api.execution.Execution`, this class here
    allows us to construct multi-objective optimization processes, i.e., such
    that have more than one optimization goal.
    """

    def __init__(self) -> None:
        """Create the multi-objective execution."""
        super().__init__()
        #: the maximum size of a pruned archive
        self._archive_max_size: int | None = None
        #: the archive size limit at which pruning should be performed
        self._archive_prune_limit: int | None = None
        #: the archive pruning strategy
        self._archive_pruner: MOArchivePruner | None = None

    def set_archive_max_size(self: TMOExecution, size: int) -> TMOExecution:
        """
        Set the upper limit for the archive size (after pruning).

        The internal archive of the multi-objective optimization process
        retains non-dominated solutions encountered during the search. Since
        there can be infinitely many such solutions, the archive could grow
        without bound if left untouched.
        Therefore, we define two size limits: the maximum archive size
        (defined by this method) and the pruning limit. Once the archive grows
        beyond the pruning limit, it is cut down to the archive size limit.

        :param size: the maximum archive size
        :returns: this execution
        """
        check_int_range(size, "maximum archive size")
        if (self._archive_prune_limit is not None) and \
                (size > self._archive_prune_limit):
            raise ValueError(
                f"archive max size {size} must be <= than archive "
                f"prune limit {self._archive_prune_limit}")
        self._archive_max_size = size
        return self

    def set_archive_pruning_limit(self: TMOExecution,
                                  limit: int) -> TMOExecution:
        """
        Set the size limit of the archive above which pruning is performed.

        If the size of the archive grows above this limit, the archive will be
        pruned down to the archive size limit.

        :param limit: the archive pruning limit
        :returns: this execution
        """
        check_int_range(limit, "limit", 1)
        if (self._archive_max_size is not None) and \
                (limit < self._archive_max_size):
            raise ValueError(
                f"archive pruning limit {limit} must be >= than archive "
                f"maximum size {self._archive_max_size}")
        self._archive_prune_limit = limit
        return self

    def set_archive_pruner(self: TMOExecution,
                           pruner: MOArchivePruner) -> TMOExecution:
        """
        Set the pruning strategy for downsizing the archive.

        :param pruner: the archive pruner
        :returns: this execution
        """
        self._archive_pruner = check_mo_archive_pruner(pruner)
        return self

    def set_objective(self: TExecution,
                      objective: Objective) -> TExecution:
        """
        Set the objective function in form of a multi-objective problem.

        :param objective: the objective function
        :returns: this execution
        """
        check_objective(objective)
        if not isinstance(objective, MOProblem):
            objective = MOSOProblemBridge(objective)
        super().set_objective(check_mo_problem(objective))
        return self

    def execute(self) -> MOProcess:
        """
        Create a multi-objective process, apply algorithm, and return result.

        This method is multi-objective equivalent of the
        :meth:`~moptipy.api.execution.Execution.execute` method. It returns a
        multi-objective process after applying the multi-objective algorithm.

        :returns: the instance of :class:`~moptipy.api.mo_process.MOProcess`
            after applying the algorithm.
        """
        objective: Final[MOProblem] = cast(MOProblem, self._objective)
        solution_space: Final[Space] = check_space(self._solution_space)
        search_space: Final[Space | None] = check_space(
            self._search_space, self._encoding is None)
        encoding: Final[Encoding | None] = check_encoding(
            self._encoding, search_space is None)
        rand_seed = self._rand_seed
        if rand_seed is not None:
            rand_seed = rand_seed_check(rand_seed)
        max_time_millis = check_max_time_millis(self._max_time_millis, True)
        max_fes = check_max_fes(self._max_fes, True)
        goal_f = check_goal_f(self._goal_f, True)
        f_lb = objective.lower_bound()
        if (f_lb is not None) and isfinite(f_lb) and \
                ((goal_f is None) or (f_lb > goal_f)):
            goal_f = f_lb

        log_all_fes = self._log_all_fes
        log_improvements = self._log_improvements or self._log_all_fes

        log_file = self._log_file
        if log_file is None:
            if log_all_fes:
                raise ValueError("Log file cannot be None "
                                 "if all FEs should be logged.")
            if log_improvements:
                raise ValueError("Log file cannot be None "
                                 "if improvements should be logged.")
        else:
            log_file.create_file_or_truncate()

        pruner: Final[MOArchivePruner] = \
            self._archive_pruner if self._archive_pruner is not None \
            else KeepFarthest(objective)
        dim: Final[int] = objective.f_dimension()
        size: Final[int] = self._archive_max_size if \
            self._archive_max_size is not None else (
            self._archive_prune_limit if
            self._archive_prune_limit is not None
            else (1 if dim == 1 else 32))
        limit: Final[int] = self._archive_prune_limit if \
            self._archive_prune_limit is not None \
            else (1 if dim == 1 else (size * 4))
        algorithm: Final[Algorithm] = check_algorithm(self._algorithm)

        process: Final[_MOProcessNoSS] = (_MOProcessNoSSLog(
            solution_space, objective, algorithm, pruner, size, limit,
            log_file, rand_seed, max_fes, max_time_millis, goal_f,
            log_all_fes) if log_improvements or log_all_fes else
            _MOProcessNoSS(solution_space, objective, algorithm, pruner,
                           size, limit, log_file, rand_seed, max_fes,
                           max_time_millis, goal_f)) \
            if search_space is None else (_MOProcessSSLog(
                solution_space, objective, algorithm, pruner, size, limit,
                log_file, search_space, encoding, rand_seed, max_fes,
                max_time_millis, goal_f,
                log_all_fes) if log_improvements or log_all_fes else
            _MOProcessSS(solution_space, objective, algorithm, pruner, size,
                         limit, log_file, search_space, encoding, rand_seed,
                         max_fes, max_time_millis, goal_f))

        try:
            # noinspection PyProtectedMember
            process._after_init()  # finalize the created process
            pruner.initialize()  # initialize the pruner
            objective.initialize()  # initialize the multi-objective problem
            if encoding is not None:
                encoding.initialize()   # initialize the encoding
            solution_space.initialize()  # initialize the solution space
            if search_space is not None:
                search_space.initialize()  # initialize the search space
            algorithm.initialize()  # initialize the algorithm
            algorithm.solve(process)  # apply the algorithm
        except Exception as be:
            # noinspection PyProtectedMember
            process._caught = be
        return process



"""
Different ways to transform and slice processes.

In this module, we provide some routines that can be used to slice of
computational budgets of a given process for running algorithms.
The following functions are included:

1. :func:`for_fes` allows for creating a sub-process that forwards all method
   calls to the original process but will perform at most a given number of
   objective function evaluations.
2. :func:`from_starting_point` creates a sub-process that has the current-best
   solution pre-set to a given point in the search space and its quality. If
   the best solution is improved upon, the provided point will be overwritten
   in place.
3. :func:`without_should_terminate` wraps a process in such a way that the
   termination criterion :meth:`~moptipy.api.process.Process.should_terminate`
   does not need to be checked anymore. Instead, once the optimization must
   stop, it will throw an internal exception and catch it again. This makes it
   possible to pass :meth:`~moptipy.api.process.Process.evaluate` to
   externally implemented algorithms that do not care about the `moptipy` API.

The utility function :func:`get_remaining_fes` returns a number representing
the remaining objective function evaluations of a given
:class:`~moptipy.api.process.Process`. If that process does not have an
FE-based termination criterion, it will instead return a very big number.
"""
import contextlib
from typing import Any, Callable, Final, TypeVar

import numpy as np
from pycommons.types import type_error

from moptipy.api.mo_process import MOProcess
from moptipy.api.process import Process, check_max_fes

#: the type variable for single- and multi-objective processes.
T = TypeVar("T", Process, MOProcess)


class __ForFEs(Process):
    """A process searching for a fixed amount of FEs."""

    def __init__(self, owner: Process, max_fes: int):
        super().__init__()
        if not isinstance(owner, Process):
            raise type_error(owner, "owner", Process)
        #: the owning process
        self.__owner: Final[Process] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        self.has_log = owner.has_log  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        self.has_best = owner.has_best  # type: ignore
        self.get_copy_of_best_x = owner.get_copy_of_best_x  # type: ignore
        self.get_best_f = owner.get_best_f  # type: ignore
        self.get_log_basename = owner.get_log_basename  # type: ignore
        #: the maximum FEs
        self.max_fes: Final[int] = check_max_fes(max_fes)
        #: the FEs that we still have left
        self.__fes_left: int = max_fes
        #: did we terminate?
        self.__terminated: bool = False
        #: the fast call to the owner's should_terminate method
        self.__should_terminate: Final[Callable[[], bool]] \
            = owner.should_terminate
        #: the fast call to the owner's evaluate method
        self.__evaluate: Final[Callable[[Any], int | float]] \
            = owner.evaluate
        #: the fast call to the owner's register method
        self.__register: Final[Callable[[Any, int | float], None]] \
            = owner.register
        #: the start fe
        self.__start_fe: Final[int] = owner.get_consumed_fes()

    def should_terminate(self) -> bool:
        return self.__terminated or self.__should_terminate()

    def terminate(self) -> None:
        self.__terminated = True

    def evaluate(self, x) -> float | int:
        f: Final[int | float] = self.__evaluate(x)
        fel: Final[int] = self.__fes_left - 1
        self.__fes_left = fel
        if fel <= 0:
            self.__terminated = True
        return f

    def register(self, x, f: int | float) -> None:
        self.__register(x, f)
        fel: Final[int] = self.__fes_left - 1
        self.__fes_left = fel
        if fel <= 0:
            self.__terminated = True

    def get_consumed_fes(self) -> int:
        return self.max_fes - self.__fes_left

    def get_last_improvement_fe(self) -> int:
        return max(1 if self.__fes_left < self.max_fes else 0,
                   self.__owner.get_last_improvement_fe() - self.__start_fe)

    def get_max_fes(self) -> int:
        return self.max_fes

    def __str__(self) -> str:
        return f"forFEs_{self.max_fes}_{self.__owner}"


class __ForFEsMO(MOProcess):
    """A process searching for a fixed amount of FEs."""

    def __init__(self, owner: MOProcess, max_fes: int):
        super().__init__()
        if not isinstance(owner, MOProcess):
            raise type_error(owner, "owner", MOProcess)
        #: the owning process
        self.__owner: Final[MOProcess] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        self.has_log = owner.has_log  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        self.has_best = owner.has_best  # type: ignore
        self.get_copy_of_best_x = owner.get_copy_of_best_x  # type: ignore
        self.get_best_f = owner.get_best_f  # type: ignore
        self.get_archive = owner.get_archive  # type: ignore
        self.check_in = owner.check_in  # type: ignore
        self.f_create = owner.f_create  # type: ignore
        self.f_validate = owner.f_validate  # type: ignore
        self.f_dtype = owner.f_dtype  # type: ignore
        self.f_dominates = owner.f_dominates  # type: ignore
        self.f_dimension = owner.f_dimension  # type: ignore
        self.get_log_basename = owner.get_log_basename  # type: ignore
        #: the maximum FEs
        self.max_fes: Final[int] = check_max_fes(max_fes)
        #: the FEs that we still have left
        self.__fes_left: int = max_fes
        #: did we terminate?
        self.__terminated: bool = False
        #: the fast call to the owner's should_terminate method
        self.__should_terminate: Final[Callable[[], bool]] \
            = owner.should_terminate
        #: the fast call to the owner's evaluate method
        self.__evaluate: Final[Callable[[Any], int | float]] \
            = owner.evaluate
        #: the fast call to the owner's register method
        self.__register: Final[Callable[[Any, int | float], None]] \
            = owner.register
        #: the evaluation wrapper
        self.__f_evaluate: Final[Callable[
            [Any, np.ndarray], int | float]] = owner.f_evaluate
        #: the start fe
        self.__start_fe: Final[int] = owner.get_consumed_fes()

    def should_terminate(self) -> bool:
        return self.__terminated or self.__should_terminate()

    def terminate(self) -> None:
        self.__terminated = True

    def evaluate(self, x) -> float | int:
        f: Final[int | float] = self.__evaluate(x)
        fel: Final[int] = self.__fes_left - 1
        self.__fes_left = fel
        if fel <= 0:
            self.__terminated = True
        return f

    def register(self, x, f: int | float) -> None:
        self.__register(x, f)
        fel: Final[int] = self.__fes_left - 1
        self.__fes_left = fel
        if fel <= 0:
            self.__terminated = True

    def f_evaluate(self, x, fs: np.ndarray) -> float | int:
        f: Final[int | float] = self.__f_evaluate(x, fs)
        fel: Final[int] = self.__fes_left - 1
        self.__fes_left = fel
        if fel <= 0:
            self.__terminated = True
        return f

    def get_consumed_fes(self) -> int:
        return self.max_fes - self.__fes_left

    def get_last_improvement_fe(self) -> int:
        return max(1 if self.__fes_left < self.max_fes else 0,
                   self.__owner.get_last_improvement_fe() - self.__start_fe)

    def get_max_fes(self) -> int:
        return self.max_fes

    def __str__(self) -> str:
        return f"forFEsMO_{self.max_fes}_{self.__owner}"


def for_fes(process: T, max_fes: int) -> T:
    """
    Create a sub-process that can run for the given number of FEs.

    :param process: the original process
    :param max_fes: the maximum number of objective function evaluations
    :returns: the sub-process that will terminate after `max_fes` FEs and that
        forwards all other calls the `process`.
    """
    max_fes = check_max_fes(max_fes, False)
    return __ForFEsMO(process, max_fes) if isinstance(process, MOProcess) \
        else __ForFEs(process, max_fes)


class __FromStartingPoint(Process):
    """A process searching from a given point."""

    def __init__(self, owner: Process, in_and_out_x: Any,
                 f: int | float):
        """
        Create a sub-process searching from one starting point.

        :param owner: the owning process
        :param in_and_out_x: the input solution record, which will be
            overwritten with the best encountered solution
        :param f: the objective value corresponding to `in_and_out`
        """
        super().__init__()
        if not isinstance(owner, Process):
            raise type_error(owner, "owner", Process)
        #: the owning process
        self.__owner: Final[Process] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        self.has_log = owner.has_log  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        self.should_terminate = owner.should_terminate  # type: ignore
        self.terminate = owner.terminate  # type: ignore
        self.get_log_basename = owner.get_log_basename  # type: ignore
        #: the best solution
        self.__best_x: Final[Any] = in_and_out_x
        #: the best-so-far solution
        self.__best_f: int | float = f
        #: the last improvement fe
        self.__last_improvement_fe: int = 0
        #: the consumed FEs
        self.__fes: int = 0
        mfes: int | None = owner.get_max_fes()
        if mfes is not None:
            mfes = mfes - owner.get_consumed_fes()
        #: the maximum permitted FEs
        self.__max_fes: Final[int | None] = mfes
        #: the fast call to the owner's evaluate method
        self.__evaluate: Final[Callable[[Any], int | float]] \
            = owner.evaluate
        #: the fast call to the owner's register method
        self.__register: Final[Callable[[Any, int | float], None]] \
            = owner.register
        #: True as long as only the seed has been used
        self.__only_seed_used: bool = True

    def has_best(self) -> bool:
        return True

    def get_copy_of_best_x(self, x) -> None:
        self.copy(x, self.__best_x)

    def get_best_f(self) -> int | float:
        return self.__best_f

    def evaluate(self, x) -> float | int:
        if self.__only_seed_used:
            if self.is_equal(x, self.__best_x):
                return self.__best_f
            self.__only_seed_used = False
        self.__fes = fe = self.__fes + 1
        f: Final[int | float] = self.__evaluate(x)
        if f <= self.__best_f:
            self.copy(self.__best_x, x)
            if f < self.__best_f:
                self.__best_f = f
                self.__last_improvement_fe = fe
        return f

    def register(self, x, f: int | float) -> None:
        if self.__only_seed_used:
            if self.is_equal(x, self.__best_x):
                return
            self.__only_seed_used = False
        self.__fes = fe = self.__fes + 1
        self.__register(x, f)
        if f < self.__best_f:
            self.copy(self.__best_x, x)
            if f < self.__best_f:
                self.__best_f = f
                self.__last_improvement_fe = fe

    def get_consumed_fes(self) -> int:
        return max(1, self.__fes)

    def get_last_improvement_fe(self) -> int:
        return self.__last_improvement_fe

    def get_max_fes(self) -> int | None:
        return self.__max_fes

    def __str__(self) -> str:
        return f"fromStart_{self.__owner}"


def from_starting_point(owner: Process, in_and_out_x: Any,
                        f: int | float) -> Process:
    """
    Create a sub-process searching from one starting point.

    This process is especially useful in conjunction with class
    :class:`~moptipy.operators.op0_forward.Op0Forward`. This class
    allows forwarding the nullary search operator to the function
    :meth:`~moptipy.api.process.Process.get_copy_of_best_x`. This way, the
    first point that it sampled by a local search can be the point specified
    as `in_and_out_x`, which effectively seeds the local search.

    To dovetail with chance of seeding, no FEs are counted at the beginning of
    the process as long as all points to be evaluated equal to the
    `in_and_out_x`. As soon as the first point different from `in_and_out_x`
    is evaluated, FE counting starts.

    Equally-good solutions will also be accepted, i.e., stored into
    `in_and_out_x`. This costs a little bit of runtime, but would normally be
    the preferred behavior: On many problems, making neutral moves (i.e.,
    drifting) will be beneficial over only accepting strict improvements. This
    is why :mod:`~moptipy.algorithms.so.rls` outperforms the normal
    :mod:`~moptipy.algorithms.so.hill_climber` on the
    :mod:`~moptipy.examples.jssp`.

    :param owner: the owning process
    :param in_and_out_x: the input solution record, which will be
        overwritten with the best encountered solution
    :param f: the objective value corresponding to `in_and_out`
    """
    return __FromStartingPoint(owner, in_and_out_x, f)


class _InternalTerminationError(Exception):
    """A protected internal termination error."""


class __WithoutShouldTerminate(Process):
    """A process allowing algorithm execution ignoring `should_terminate`."""

    def __init__(self, owner: Process):
        super().__init__()
        if not isinstance(owner, Process):
            raise type_error(owner, "owner", Process)
        #: the owning process
        self._owner: Final[Process] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        a = owner.get_last_improvement_fe  # type: ignore
        self.get_last_improvement_fe = a  # type: ignore
        self.has_log = owner.has_log  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        self.has_best = owner.has_best  # type: ignore
        self.get_copy_of_best_x = owner.get_copy_of_best_x  # type: ignore
        self.get_best_f = owner.get_best_f  # type: ignore
        self.should_terminate = owner.should_terminate  # type: ignore
        self.terminate = owner.terminate  # type: ignore
        self.get_max_fes = owner.get_max_fes  # type: ignore
        self.get_consumed_fes = owner.get_consumed_fes  # type: ignore
        self.get_log_basename = owner.get_log_basename  # type: ignore
        #: the fast call to the owner's evaluate method
        self.__evaluate: Final[Callable[[Any], int | float]] \
            = owner.evaluate
        #: the fast call to the owner's register method
        self.__register: Final[Callable[[Any, int | float], None]] \
            = owner.register

    def evaluate(self, x) -> float | int:
        if self.should_terminate():
            raise _InternalTerminationError
        return self.__evaluate(x)

    def register(self, x, f: int | float) -> None:
        if self.should_terminate():
            raise _InternalTerminationError
        self.__register(x, f)

    def __str__(self) -> str:
        return f"protect_{self._owner}"

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        return True


class __WithoutShouldTerminateMO(MOProcess):
    """A process allowing algorithm execution ignoring `should_terminate`."""

    def __init__(self, owner: Process):
        super().__init__()
        if not isinstance(owner, MOProcess):
            raise type_error(owner, "owner", MOProcess)
        #: the owning process
        self._owner: Final[MOProcess] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        a = owner.get_last_improvement_fe  # type: ignore
        self.get_last_improvement_fe = a  # type: ignore
        self.has_log = owner.has_log  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        self.has_best = owner.has_best  # type: ignore
        self.get_copy_of_best_x = owner.get_copy_of_best_x  # type: ignore
        self.get_best_f = owner.get_best_f  # type: ignore
        self.should_terminate = owner.should_terminate  # type: ignore
        self.terminate = owner.terminate  # type: ignore
        self.get_max_fes = owner.get_max_fes  # type: ignore
        self.get_consumed_fes = owner.get_consumed_fes  # type: ignore
        self.f_create = owner.f_create  # type: ignore
        self.f_validate = owner.f_validate  # type: ignore
        self.f_dtype = owner.f_dtype  # type: ignore
        self.f_dominates = owner.f_dominates  # type: ignore
        self.f_dimension = owner.f_dimension  # type: ignore
        self.get_archive = owner.get_archive  # type: ignore
        self.check_in = owner.check_in  # type: ignore
        self.get_log_basename = owner.get_log_basename  # type: ignore
        #: the fast call to the owner's evaluate method
        self.__evaluate: Final[Callable[[Any], int | float]] \
            = owner.evaluate
        #: the fast call to the owner's register method
        self.__register: Final[Callable[[Any, int | float], None]] \
            = owner.register
        #: the fast call to the owner's f_evaluate method
        self.__f_evaluate: Final[Callable[
            [Any, np.ndarray], int | float]] \
            = owner.f_evaluate

    def evaluate(self, x) -> float | int:
        if self.should_terminate():
            raise _InternalTerminationError
        return self.__evaluate(x)

    def register(self, x, f: int | float) -> None:
        if self.should_terminate():
            raise _InternalTerminationError
        self.__register(x, f)

    def f_evaluate(self, x, fs: np.ndarray) -> int | float:
        if self.should_terminate():
            raise _InternalTerminationError
        return self.__f_evaluate(x, fs)

    def __str__(self) -> str:
        return f"protectMO_{self._owner}"

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        return True


def get_remaining_fes(process: Process) -> int:
    """
    Get a finite number representing the remaining FEs of a process.

    If the process has the maximum objective function evaluations (FEs) set
    (see :meth:`~moptipy.api.process.Process.get_max_fes`), then this method
    returns the maximum FEs minus the consumed FEs (see
    :meth:`~moptipy.api.process.Process.get_consumed_fes`).
    Otherwise, i.e., if :meth:`~moptipy.api.process.Process.get_max_fes`
    returns `None`, this function returns a very large number, namely
    `9223372036854775807`, i.e., `(2 ** 63) - 1`. This number is so high that
    it will always be impossible to consume it in terms of FEs. But it is
    also finite in any case. When trying to slice of budgets or computing
    things based on the remaining budget, this makes it unnecessary for us to
    deal with special cases.

    :param process: the process
    :returns: an integer representing the remaining FEs of the process. If
        no FE limit is imposed by `process`, a very large number will be
        returned.

    >>> from moptipy.api.process import Process as Proc
    >>> class X(Proc):
    ...     def get_max_fes(self):
    ...         return None
    ...     def get_consumed_fes(self):
    ...         return 123
    >>> get_remaining_fes(X())
    9223372036854775807
    >>> class Y(X):
    ...     def get_max_fes(self):
    ...         return 456
    >>> get_remaining_fes(Y())
    333
    """
    mf: int | None = process.get_max_fes()  # get the number of available FEs
    if mf is None:  # if a no FE limit is specified, then return a large value
        return 9_223_372_036_854_775_807  # (2 ** 63) - 1
    return mf - process.get_consumed_fes()  # else, subtract the consumed FEs


def without_should_terminate(algorithm: Callable[[T], Any], process: T) \
        -> None:
    """
    Apply an algorithm that does not call `should_terminate` to a process.

    If we use an algorithm from an external library, this algorithm may ignore
    the proper usage of our API. With this method, we try to find a way to
    make sure that these calls are consistent with the termination criterion
    of the `moptipy` API.

    Before calling :meth:`~moptipy.api.process.Process.evaluate`,
    :meth:`~moptipy.api.process.Process.register`, or
    :meth:`~moptipy.api.mo_problem.MOProblem.f_evaluate`, an optimization
    algorithm must check if it should instead stop via
    :meth:`~moptipy.api.process.Process.should_terminate`. If the process
    called :meth:`~moptipy.api.process.Process.should_terminate` and was
    told to stop but did invoke the evaluation routines anyway, an
    exception will be thrown and the process force-terminates. If the
    process did not call
    :meth:`~moptipy.api.process.Process.should_terminate` but was
    supposed to stop, the results of
    :meth:`~moptipy.api.process.Process.evaluate` may be arbitrary (or
    positive infinity).

    This function here can be used to deal with processes that do not invoke
    :meth:`~moptipy.api.process.Process.should_terminate`. It will invoke
    this method by itself before
    :meth:`~moptipy.api.process.Process.evaluate`,
    :meth:`~moptipy.api.process.Process.register`, and
    :meth:`~moptipy.api.mo_problem.MOProblem.f_evaluate` and terminate the
    algorithm with an exception if necessary. It will then catch the
    exception and bury it.

    Thus, we can now use algorithms that ignore our termination criteria and
    still force them to terminate when they should.

    Some algorithms using this system are implemented in
    :mod:`~moptipy.algorithms.so.vector.scipy` and
    :mod:`~moptipy.algorithms.so.vector.pdfo`. These modules import external
    algorithms from other libraries which, of course, know nothing about how
    our `moptipy` works. They only accept the objective function and cannot
    handle the beautiful
    :meth:`~moptipy.api.process.Process.should_terminate`-based termination
    criteria. By using :func:`without_should_terminate`, however, we can still
    safely use them within `moptipy` compliant scenarios.

    :param algorithm: the algorithm
    :param process: the optimization process
    """
    with contextlib.suppress(_InternalTerminationError), \
            __WithoutShouldTerminateMO(process) \
            if isinstance(process, MOProcess) \
            else __WithoutShouldTerminate(process) as proc:
        algorithm(proc)



"""
In this package, some pre-defined search and solution spaces are provided.

These spaces implement the base class :class:`~moptipy.api.space.Space`, which
defines the API for creating, copying, storing, and comparing (for equality)
of data structures that can be used to represent either points in the search
space or candidate solutions in the solution space.

The following pre-defined spaces are currently available:

- :class:`~moptipy.spaces.bitstrings.BitStrings`, the space of `n`-dimensional
  bit strings
- :class:`~moptipy.spaces.intspace.IntSpace`, a space of `n`-dimensional
  integer strings, where each element is between (and including) a minimum
  and a maximum value (inclusive)
- :class:`~moptipy.spaces.permutations.Permutations` is a special version of
  the :class:`~moptipy.spaces.intspace.IntSpace` where all elements are
  permutations of a base string
  :attr:`~moptipy.spaces.permutations.Permutations.blueprint`. This means that
  it can represent permutations both with and without repetitions. Depending
  on the base string, each element may occur an element-specific number of
  times. For the base string `(-1, -1, 2, 7, 7, 7)`, for example, `-1` may
  occur twice, `2` can occur once, and `7` three times.
- :class:`~moptipy.spaces.ordered_choices.OrderedChoices` is a combination of
  permutations and combinations. There are `n` choices of one or multiple
  different values each. The choices are either disjoint or identical. An
  element from the space picks one value per choice. The order of the elements
  matters.
- :class:`~moptipy.spaces.vectorspace.VectorSpace` is the space of
  `n`-dimensional floating point number vectors
"""



"""The mock multi-objective optimization problem."""

from typing import Iterable, cast

import numpy as np
from numpy.random import default_rng
from pycommons.types import check_int_range, type_error

from moptipy.mo.problem.weighted_sum import WeightedSum
from moptipy.mock.objective import MockObjective


class MockMOProblem(WeightedSum):
    """A mock-up of a multi-objective optimization problem."""

    def __init__(self, objectives: Iterable[MockObjective],
                 weights: Iterable[int | float]) -> None:
        """
        Create the mock multi-objective problem.

        :param objectives: the mock objectives
        :param weights: their weights
        """
        super().__init__(objectives, weights)

    @staticmethod
    def for_dtype(n: int, dtype: np.dtype) -> "MockMOProblem":
        """
        Create a mock multi-objective problem.

        :param n: the number of objectives
        :param dtype: the optional dtype
        :returns: the mock multi-objective problem
        """
        check_int_range(n, "n", 1, 1_000)
        if not isinstance(dtype, np.dtype):
            raise type_error(dtype, "dtype", np.dtype)

        random = default_rng()
        weights: list[int | float] =\
            [int(w) for w in random.integers(1, 3, n)] \
            if random.integers(2) <= 0 \
            else [float(w) for w in random.uniform(0.2, 3, n)]
        max_trials: int = 1000
        while max_trials > 0:
            max_trials -= 1
            funcs = [MockObjective.for_type(dtype) for _ in range(n)]
            names = {str(o) for o in funcs}
            if len(names) >= n:
                return MockMOProblem(funcs, weights)
        raise ValueError("Huh?")

    def get_objectives(self) -> tuple[MockObjective, ...]:
        """
        Get the internal objective functions.

        :return: the internal mock objective functions
        """
        return cast(tuple[MockObjective, ...], self._objectives)

    def sample(self, fs: np.ndarray) -> int | float:
        """
        Sample one vector of ojective values.

        :param fs: the array to receive the objective values
        :returns: the scalarized objective values
        """
        for i, o in enumerate(self._objectives):
            fs[i] = cast(MockObjective, o).sample()
        return self._scalarize(fs)



"""A nullary operator filling a vector with uniformly distribute values."""

from typing import Final

import numpy as np
from numpy.random import Generator
from pycommons.types import type_error

from moptipy.api.operators import Op0
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.logger import KeyValueLogSection


class Op0Uniform(Op0):
    """Fill a vector with uniformly distributed random values."""

    def __init__(self, space: VectorSpace) -> None:
        """
        Initialize the operator.

        :param space: the vector space
        """
        super().__init__()
        if not isinstance(space, VectorSpace):
            raise type_error(space, "space", VectorSpace)
        #: store the space
        self.space: Final[VectorSpace] = space

    def __str__(self) -> str:
        """
        Get the name of this operator.

        :returns: the name of this space
        """
        return "uniform"

    def op0(self, random: Generator, dest: np.ndarray) -> None:
        """
        Fill the string `dest` with random values.

        :param random: the random number generator
        :param dest: the vector to be filled. Afterwards it contains uniformly
            distributed random values
        """
        sp: Final[VectorSpace] = self.space
        np.copyto(dest, random.uniform(
            sp.lower_bound, sp.upper_bound, sp.dimension))

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this operator to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        self.space.log_bounds(logger)



"""Default styles for plots."""
from typing import Final, cast

import matplotlib.cm as mplcm  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import numpy as np
from matplotlib import colors  # type: ignore
from pycommons.types import check_int_range, type_error

#: The internal color black.
COLOR_BLACK: Final[tuple[float, float, float]] = (0.0, 0.0, 0.0)
#: The internal color white.
COLOR_WHITE: Final[tuple[float, float, float]] = (1.0, 1.0, 1.0)


def str_to_palette(palette: str) \
        -> tuple[tuple[float, float, float], ...]:
    """
    Obtain a palette from a string.

    :param palette: the string with all the color data.
    :returns: the palette
    """
    if isinstance(colors, str):
        raise type_error(colors, "colors", str)
    result: list[tuple[float, float, float]] = []
    end: int = -1
    length: Final[int] = len(palette)
    while end < length:
        start: int = end + 1
        end = length
        for ch in "\n\r\t ,;":
            end_new: int = palette.find(ch, start)
            if start < end_new < end:
                end = end_new
        color: str = palette[start:end].strip()
        if color.startswith("#"):
            color = color[1:].strip()
        if len(color) > 0:
            if len(color) != 6:
                raise ValueError(f"invalid color: {color!r} in {palette!r}")
            result.append((int(color[0:2], 16) / 255,
                           int(color[2:4], 16) / 255,
                           int(color[4:6], 16) / 255))
    if len(result) <= 0:
        raise ValueError(f"empty colors: {palette!r}")
    return tuple(result)


#: A palette with 11 distinct colors.
__PALETTE_11: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "ff0000 0000ff 00d900 a32626 ffa500 008833 7788ff cf00ff 770066 cccc00 "
    "a0a9a0")

#: A palette with 21 distinct colors.
__PALETTE_21: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "e6194b 3cb44b ffe119 0082c8 f58231 911eb4 46f0f0 f032e6 d2f53c fabebe "
    "008080 e6beff aa6e28 cfca08 800000 aaffc3 808000 ffd8b1 000080 505000 "
    "90a0a0")

#: A palette with 25 distinct colors.
__PALETTE_25: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "20b2aa f0e68c b8860b ff00ff 00fa9a ee82ee 9acd32 dcdcdc 556b2f db7093 "
    "00bfff 1e90ff 8a2be2 ff1493 008000 ffd700 000080 800000 7fff00 8b008b "
    "0000ff 483d8b ff0000 696969 ff7f50")

#: A palette with 28 distinct colors.
__PALETTE_28: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "545454 550000 b0b0ff bababa 878787 005500 550055 b000b0 e4e400 00b0b0 "
    "00ffff 00ff00 878700 870087 005555 ff0000 0000ff baba00 00b000 ff00ff "
    "870000 4949ff 008700 8484ff 008787 e4e4e4 545400 b00000")

#: A palette with 30 distinct colors.
__PALETTE_30: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "00bfff ff1493 7f007f 8a2be2 1e90ff adff2f ffb6c1 fa8072 ffff54 90ee90 "
    "ff4500 00ff00 483d8b ffa500 696969 add8e6 ff00ff 7f0000 8fbc8f da70d6 "
    "006400 dc143c 00ff7f 008b8b 808000 b03060 00ffff 0000ff 000080 deb887 ")

#: A palette with 35 distinct colors.
__PALETTE_35: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "228b22 808080 9acd32 008080 483d8b 7f0000 d2b48c 00ff00 00008b 808000 "
    "00ced1 8fbc8f 556b2f ffc0cb 1e90ff ffff54 800080 00bfff ff00ff 7b68ee "
    "ff6347 9400d3 ffa500 7fffd4 b03060 ff0000 f08080 f4a460 00ff7f 98fb98 "
    "f0e68c ff1493 0000ff b0e0e6 dda0dd")

#: A palette with 40 distinct colors.
__PALETTE_40: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "ff00ff 7f007f a9a9a9 0000ff daa520 2f4f4f da70d6 ff8c00 ffc0cb afeeee "
    "00008b 20b2aa 00ff00 f4a460 8b0000 008000 f08080 a0522d ff6347 808000 "
    "bdb76b 98fb98 191970 ffe4c4 ffff00 4682b4 32cd32 1e90ff 3cb371 ff1493 "
    "b03060 7fffd4 556b2f dda0dd adff2f 7b68ee 9acd32 ff0000 00bfff 9400d3 ")

#: A palette with 45 distinct colors.
__PALETTE_45: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "ffb60c 00008b 8b4513 e9967a a020f0 ff0000 6a5acd 808080 008b8b 00ff00 "
    "90ee90 4682b4 daa520 7f007f ffd700 00bfff 00ced1 556b2f c0c0c0 008000 "
    "ff69b4 dda0dd adff2f 3cb371 8b0000 0000ff f5deb3 32cd32 1e90ff b03060 "
    "ff7f50 cd5c5c ff00ff 7fffd4 6b8e23 dc143c 9acd32 afeeee 2f4f4f 00fa9a "
    "191970 8fbc8f ba55d3 ff8c00 bdb76b")

#: A palette with 50 distinct colors.
__PALETTE_50: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "7fffd4 000080 00ff7f 483d8b 008000 afeeee ffc0cb daa520 ffd700 ff8c00 "
    "87cefa ff00ff c0c0c0 cd5c5c 00ff00 adff2f e9967a ffff00 dc143c cd853f "
    "9932cc 98fb98 708090 800000 ee82ee 800080 8fbc8f 4682b4 dda0dd 32cd32 "
    "a020f0 48d1cc a0522d ff1493 ff69b4 f0e68c 9370db 2f4f4f ff0000 6b8e23 "
    "556b2f ff6347 f5deb3 008b8b db7093 0000cd 9acd32 3cb371 0000ff 6495ed")

#: A palette with 60 distinct colors.
__PALETTE_60: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "d2b48c 191970 000080 20b2aa bc8f8f 2f4f4f 8b4513 ff4500 dc143c afeeee "
    "db7093 c71585 bdb76b b03060 708090 00ffff 6495ed 808000 f08080 b8860b "
    "cd853f 556b2f ffa07a 008080 eee8aa dda0dd ee82ee 7fffd4 9370db 7f007f "
    "d8bfd8 9acd32 4682b4 98fb98 00ff00 9932cc 0000cd d2691e 00fa9a ff8c00 "
    "663399 a9a9a9 ffff00 32cd32 8b0000 4169e1 8fbc8f 0000ff ff69b4 008000 "
    "ffb6c1 00bfff ffd700 ff1493 ff00ff 3cb371 a020f0 87ceeb ff6347 adff2f")

#: A palette with 70 distinct colors.
__PALETTE_70: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "b8860b a52a2a ffe4c4 663399 da70d6 a020f0 eee8aa 9acd32 ba55d3 ff8c00 "
    "00008b 2e8b57 66cdaa 2f4f4f 0000cd 32cd32 556b2f adff2f 7fffd4 00ff00 "
    "008b8b 4b0082 ffd700 00bfff 6b8e23 48d1cc b03060 778899 ffff54 ff0000 "
    "00ff7f d2b48c 3cb371 bc8f8f b0e0e6 ffc0cb dcdcdc ff00ff cd5c5c 6495ed "
    "0000ff 9370db dc143c a9a9a9 bdb76b f4a460 c71585 e9967a 4682b4 db7093 "
    "483d8b 8b008b 87cefa 7fff00 dda0dd 006400 9932cc 808000 ff1493 191970 "
    "4169e1 ff6347 b0c4de ff69b4 d2691e 00ffff 98fb98 ffff00 a0522d 8fbc8f")

#: A palette with 80 distinct colors.
__PALETTE_80: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "eee8aa bdb76b 2e8b57 0000ff 1e90ff cd5c5c ffd700 90ee90 00ffff ba55d3 "
    "d3d3d3 40e0d0 ff0000 b0c4de 008000 0000cd 00bfff ffa07a ff00ff 778899 "
    "dc143c cd853f 556b2f 9acd32 9400d3 663399 b03060 ffdab9 b22222 ffa500 "
    "7fffd4 4169e1 dda0dd 8fbc8f ffff00 ff1493 191970 00008b 9932cc 006400 "
    "483d8b 808000 696969 00ff7f d8bfd8 d2691e 4682b4 a9a9a9 c71585 5f9ea0 "
    "ffc0cb 7f0000 800080 ee82ee 008080 4b0082 ffff54 9370db 8b4513 ff6347 "
    "daa520 87cefa adff2f ff8c00 6b8e23 deb887 00fa9a 66cdaa 7b68ee fa8072 "
    "ff69b4 db7093 b0e0e6 00ff00 2f4f4f bc8f8f 32cd32 3cb371 20b2aa 7cfc00")

#: A palette with 90 distinct colors.
__PALETTE_90: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "fa8072 ffff54 0000cd 32cd32 dc143c 40e0d0 ff8c00 9932cc 7b68ee 800000 "
    "4b0082 008000 afeeee ee82ee ffd700 6a5acd 2e8b57 eee8aa a9a9a9 ff69b4 "
    "b8860b c71585 ff4500 7fff00 2f4f4f 00ff7f ffa500 add8e6 008b8b bdb76b "
    "ff00ff cd853f f0e68c 808080 db7093 800080 a0522d 87ceeb 3cb371 9400d3 "
    "b0c4de daa520 b03060 dda0dd 66cdaa 6b8e23 d3d3d3 4682b4 9acd32 4169e1 "
    "ff7f50 5f9ea0 f08080 00fa9a 00ced1 808000 adff2f 0000ff d2691e ff1493 "
    "ff6347 ffdead 20b2aa 006400 7fffd4 483d8b cd5c5c f4a460 8fbc8f 663399 "
    "d8bfd8 9370db b22222 d2b48c ffff00 ba55d3 bc8f8f ffb6c1 6495ed e9967a "
    "ff0000 000080 00ff00 191970 90ee90 00ffff 778899 1e90ff 00bfff 556b2f")

#: A palette with 100 distinct colors.
__PALETTE_100: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "bdb76b 8a2be2 bc8f8f eee8aa 8fbc8f a020f0 ff4500 2f4f4f 000080 ff6347 "
    "6b8e23 808080 663399 afeeee dcdcdc ff8c00 3cb371 191970 4682b4 7cfc00 "
    "ee82ee 32cd32 9370db ff1493 00ff00 6a5acd adff2f 006400 00fa9a ffd700 "
    "d2691e c0c0c0 00ff7f ba55d3 add8e6 ffff00 6495ed 8b008b 48d1cc 87cefa "
    "fa8072 deb887 b22222 7b68ee b8860b 90ee90 8b4513 9acd32 cd853f f08080 "
    "ff00ff da70d6 9400d3 ff69b4 ffe4b5 556b2f 8b0000 1e90ff e9967a 696969 "
    "c71585 ffa07a f0e68c a9a9a9 20b2aa a0522d 87ceeb 0000ff 008000 daa520 "
    "00ffff b03060 483d8b ffb6c1 dda0dd ff0000 9932cc 708090 cd5c5c 4b0082 "
    "ffa500 0000cd dc143c a52a2a 008080 7fffd4 5f9ea0 b0c4de d8bfd8 808000 "
    "db7093 f4a460 2e8b57 ffff54 40e0d0 66cdaa ffe4c4 ff7f50 4169e1 00bfff")

#: A palette with 110 distinct colors.
__PALETTE_110: Final[tuple[tuple[float, float, float], ...]] = str_to_palette(
    "0000ff 556b2f ff4500 000080 f5deb3 ff1493 00008b 0000cd ff8c00 ffc0cb "
    "fa8072 696969 da70d6 8b4513 d2691e 00fa9a 008080 c71585 bdb76b 40e0d0 "
    "6495ed daa520 5f9ea0 ffe4c4 f4a460 ba55d3 228b22 00ced1 00ffff 66cdaa "
    "b03060 9400d3 f08080 dc143c adff2f 00ff00 9932cc ee82ee 8a2be2 32cd32 "
    "afeeee 191970 808080 9acd32 90ee90 b22222 a0522d ff00ff 006400 c0c0c0 "
    "2f4f4f d3d3d3 cd853f 98fb98 9370db 4169e1 4b0082 00ff7f ff7f50 8fbc8f "
    "8b0000 eee8aa dda0dd 008000 ffa500 ff6347 ff69b4 d8bfd8 ffa07a 4682b4 "
    "b0c4de 2e8b57 ff0000 00bfff f0e68c 7f0000 b8860b d2b48c ffff54 1e90ff "
    "e9967a 3cb371 87cefa 87ceeb 6a5acd ffd700 708090 bc8f8f 7b68ee 6b8e23 "
    "a52a2a 7fffd4 ffdead ffdab9 808000 483d8b 20b2aa a020f0 48d1cc db7093 "
    "a9a9a9 663399 b0e0e6 deb887 7f007f 7fff00 8b008b ffff00 cd5c5c add8e6")


#: A set of predefined uniquely-looking colors.
__FIXED_COLORS: Final[tuple[tuple[tuple[float, float, float], ...], ...]] = (
    __PALETTE_11, __PALETTE_21, __PALETTE_25, __PALETTE_28, __PALETTE_30,
    __PALETTE_35, __PALETTE_40, __PALETTE_45, __PALETTE_50, __PALETTE_60,
    __PALETTE_70, __PALETTE_80, __PALETTE_90, __PALETTE_100, __PALETTE_110)


def distinct_colors(n: int) -> tuple[tuple[float, float, float], ...]:
    """
    Obtain a set of `n` distinct colors.

    :param n: the number of colors required
    :return: a tuple of colors
    """
    check_int_range(n, "n", 1, 1000)

    # First, let us see if we can cover the range with hand-picked colors.
    for k in __FIXED_COLORS:
        lk = len(k)
        if lk >= n:
            if lk == n:
                return k
            return tuple(k[0:n])

    # Second, let's see whether the method from
    # https://stackoverflow.com/questions/8389636
    # works.
    # This method does not seem to make good use of the available color space.
    # Since we use it only for cases with more than 110 colors, that's OK.
    cm = plt.get_cmap("gist_rainbow")
    c_norm = colors.Normalize(vmin=0, vmax=n - 1)
    scalar_map = mplcm.ScalarMappable(norm=c_norm, cmap=cm)
    qq = cast(list[tuple[float, float, float]],
              [tuple(scalar_map.to_rgba(i)[0:3]) for i in np.arange(n)])
    ss = set(qq)
    if len(ss) == n:
        return tuple(qq)

    raise ValueError(f"Could not obtain {n} distinct colors.")


#: The solid line dash
LINE_DASH_SOLID: Final[str] = "solid"

#: An internal array of fixed line styles.
__FIXED_LINE_DASHES: \
    Final[tuple[str | tuple[float, tuple[float, ...]], ...]] = \
    (LINE_DASH_SOLID,
     "dashed",
     "dashdot",
     "dotted",
     (0.0, (3.0, 5.0, 1.0, 5.0, 1.0, 5.0)),  # dashdotdotted
     (0.0, (3.0, 1.0, 1.0, 1.0)),  # densely dashdotted
     (0.0, (5.0, 1.0)),  # densely dashed
     (0.0, (1.0, 1.0)),  # densely dotted
     (0.0, (3.0, 1.0, 1.0, 1.0, 1.0, 1.0)),  # densely dashdotdotted
     (0.0, (1.0, 10.0)),  # loosely dotted
     (0.0, (5.0, 10.0)),  # loosely dashed
     (0.0, (3.0, 10.0, 1.0, 10.0)),  # loosely dashdotted
     (0.0, (3.0, 10.0, 1.0, 10.0, 1.0, 10.0)))  # loosely dashdotdotted


def distinct_line_dashes(n: int) -> \
        tuple[str | tuple[float, tuple[float, ...]], ...]:
    """
    Create a sequence of distinct line dashes.

    :param n: the number of styles
    :return: the styles
    """
    check_int_range(n, "n", 1, len(__FIXED_LINE_DASHES) - 1)
    if n == __FIXED_LINE_DASHES:
        return __FIXED_LINE_DASHES
    return tuple(__FIXED_LINE_DASHES[0:n])


#: The fixed predefined distinct markers
__FIXED_MARKERS: tuple[str, ...] = ("o", "^", "s", "P", "X", "D", "*", "p")


def distinct_markers(n: int) -> tuple[str, ...]:
    """
    Create a sequence of distinct markers.

    :param n: the number of markers
    :return: the markers
    """
    lfm: Final[int] = len(__FIXED_MARKERS)
    check_int_range(n, "n", 1, lfm)
    if n == lfm:
        return __FIXED_MARKERS
    return tuple(__FIXED_MARKERS[0:n])


def importance_to_line_width(importance: int) -> float:
    """
    Transform an importance value to a line width.

    Basically, an importance of `0` indicates a normal line in a normal
    plot that does not need to be emphasized.
    A positive importance means that the line should be emphasized.
    A negative importance means that the line should be de-emphasized.

    :param importance: a value between -9 and 9
    :return: the line width
    """
    check_int_range(importance, "importance", -9, 9)
    if importance >= 0:
        return 2.0 * (0.5 + importance)
    if importance == -1:
        return 2.0 / 3.0
    if importance == -2:
        return 0.5
    return 0.7 ** (-importance)


def importance_to_alpha(importance: int) -> float:
    """
    Transform an importance value to an alpha value.

    Basically, an importance of `0` indicates a normal line in a normal
    plot that does not need to be emphasized.
    A positive importance means that the line should be emphasized.
    A negative importance means that the line should be de-emphasized.

    :param importance: a value between -9 and 9
    :return: the alpha
    """
    check_int_range(importance, "importance", -9, 9)
    if importance >= 0:
        return 1.0
    if importance == -1:
        return 2.0 / 3.0
    if importance == -2:
        return 0.5
    return 1.0 / 3.0


#: The internal default basic style
__BASE_LINE_STYLE: Final[dict[str, object]] = {
    "alpha": 1.0,
    "antialiased": True,
    "color": COLOR_BLACK,
    "dash_capstyle": "butt",
    "dash_joinstyle": "round",
    "linestyle": LINE_DASH_SOLID,
    "linewidth": 1.0,
    "solid_capstyle": "round",
    "solid_joinstyle": "round",
}


def create_line_style(**kwargs) -> dict[str, object]:
    """
    Obtain the basic style for lines in diagrams.

    :param kwargs: any additional overrides
    :return: a dictionary with the style elements
    """
    res = dict(__BASE_LINE_STYLE)
    res.update(kwargs)
    return res


def importance_to_font_size(importance: float) -> float:
    """
    Transform an importance value to a font size.

    :param importance: the importance value
    :return: the font size
    """
    check_int_range(importance, "importance", -9, 9)
    if importance < 0:
        return 7.5
    if importance <= 0:
        return 8.0
    if importance == 1:
        return 8.5
    if importance == 2:
        return 9.0
    if importance == 3:
        return 10.0
    return 11.0


#: The default grid color
GRID_COLOR: Final[tuple[float, float, float]] = \
    (7.0 / 11.0, 7.0 / 11.0, 7.0 / 11.0)


def rgb_to_gray(r: float, g: float, b: float) -> float:
    """
    Convert RGB values to gray scale.

    :param r: the red value
    :param g: the green value
    :param b: the blue value
    :return: the gray value
    """
    return (0.2989 * r) + (0.5870 * g) + (0.1140 * b)


def text_color_for_background(background: tuple[float, float, float]) \
        -> tuple[float, float, float]:
    """
    Get a reasonable text color for a given background color.

    :param background: the background color
    :return: the text color
    """
    br: Final[float] = background[0]
    bg: Final[float] = background[1]
    bb: Final[float] = background[2]
    bgg: Final[float] = rgb_to_gray(br, bg, bb)

    return COLOR_WHITE if bgg < 0.3 else COLOR_BLACK



"""
An implementation of the operation-based encoding for JSSPs.

1. Mitsuo Gen, Yasuhiro Tsujimura, and Erika Kubota. Solving Job-Shop
   Scheduling Problems by Genetic Algorithm. In *Humans, Information and
   Technology: Proceedings of the 1994 IEEE International Conference on
   Systems, Man and Cybernetics,* October 2-5, 1994, San Antonio, TX, USA,
   volume 2. Piscataway, NJ, USA: IEEE. ISBN: 0-7803-2129-4.
   doi: https://doi.org/10.1109/ICSMC.1994.400072.
   http://read.pudn.com/downloads151/doc/658565/00400072.pdf.
2. Christian Bierwirth. A Generalized Permutation Approach to Job Shop
   Scheduling with Genetic Algorithms. *Operations-Research-Spektrum
   (OR Spectrum)*, 17(2-3):87-92, June 1995.
   doi: https://doi.org/10.1007/BF01719250.
   https://www.researchgate.net/publication/240263036.
3. Christian Bierwirth, Dirk C. Mattfeld, and Herbert Kopfer. On Permutation
   Representations for Scheduling Problems. In Hans-Michael Voigt, Werner
   Ebeling, Ingo Rechenberg, and Hans-Paul Schwefel, editors, *Proceedings of
   the 4th International Conference on Parallel Problem Solving from Nature
   (PPSN IV),* September 22-24, 1996, Berlin, Germany, pages 310-318.
   Volume 1141/1996 of Lecture Notes in Computer Science (LNCS), Berlin,
   Germany: Springer-Verlag GmbH. ISBN: 3-540-61723-X.
   doi: https://doi.org/10.1007/3-540-61723-X_995.
   https://www.researchgate.net/publication/2753293.
4. Guoyong Shi, Hitoshi Iima, and Nobuo Sannomiya. New Encoding Scheme for
   Solving Job Shop Problems by Genetic Algorithm. In *Proceedings of the 35th
   IEEE Conference on Decision and Control (CDC'96),* December 11-13, 1996,
   Kobe, Japan, volume 4, pages 4395-4400. Piscataway, NJ, USA: IEEE.
   ISBN: 0-7803-3590-2. doi: https://doi.org/10.1109/CDC.1996.577484.
   https://www.researchgate.net/publication/224238934.
"""
from typing import Final

import numba  # type: ignore
import numpy as np
from pycommons.types import type_error

from moptipy.api.encoding import Encoding
from moptipy.examples.jssp.instance import Instance
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import (
    KEY_NUMPY_TYPE,
    int_range_to_dtype,
    numpy_type_to_str,
)

#: the numpy data type for machine indices
KEY_NUMPY_TYPE_MACHINE_IDX: Final[str] = f"{KEY_NUMPY_TYPE}MachineIdx"
#: the numpy data type for job indices
KEY_NUMPY_TYPE_JOB_IDX: Final[str] = f"{KEY_NUMPY_TYPE}JobIdx"
#: the numpy data type for job times
KEY_NUMPY_TYPE_JOB_TIME: Final[str] = f"{KEY_NUMPY_TYPE}JobTime"


# start book
@numba.njit(nogil=True, cache=True)
def decode(x: np.ndarray, machine_idx: np.ndarray,
           job_time: np.ndarray, job_idx: np.ndarray,
           instance: np.ndarray, y: np.ndarray) -> None:
    """
    Map an operation-based encoded array to a Gantt chart.

    :param x: the source array, i.e., multi-permutation
    :param machine_idx: array of length `m` for machine indices
    :param job_time: array of length `n` for job times
    :param job_idx: length `n` array of current job operations
    :param instance: the instance data matrix
    :param y: the output array, i.e., the Gantt chart
    """
    machine_idx.fill(-1)  # all machines start by having done no jobs
    job_time.fill(0)  # each job has initially consumed 0 time units
    job_idx.fill(0)  # each job starts at its first operation

    for job in x:  # iterate over multi-permutation
        idx = job_idx[job]  # get the current operation of the job
        job_idx[job] = idx + 1  # and step it to the next operation
        machine = instance[job, idx, 0]  # get the machine id
        start = job_time[job]  # end time of previous operation of job
        mi = machine_idx[machine]  # get jobs finished on machine - 1
        if mi >= 0:  # we already have one job done?
            start = max(start, y[machine, mi, 2])  # earliest start
        mi += 1  # step the machine index
        machine_idx[machine] = mi  # step the machine index
        end = start + instance[job, idx, 1]  # compute end time
        y[machine, mi, 0] = job  # store job index
        y[machine, mi, 1] = start  # store start of job's operation
        y[machine, mi, 2] = end  # store end of job's operation
        job_time[job] = end  # time next operation of job can start
    # end book


# start book
class OperationBasedEncoding(Encoding):
    # reusable variables __machine_time, __job_time, and __job_idx are
    # allocated in __init__; __matrix refers to instance data matrix
    # end book
    """
    An operation-based encoding for the Job Shop Scheduling Problem (JSSP).

    The operation-based encoding for the Job Shop Scheduling Problem (JSSP)
    maps permutations with repetitions to Gantt charts.
    In the operation-based encoding, the search space consists of permutations
    with repetitions of length `n*m`, where `n` is the number of jobs in the
    JSSP instance and `m` is the number of machines.
    In such a permutation with repetitions, each of the job ids from `0..n-1`
    occurs exactly `m` times.
    In the encoding, the permutations are processed from beginning to end and
    the jobs are assigned to machines in exactly the order in which they are
    encountered. If a job is encountered for the first time, we place its
    first operation onto the corresponding machine. If we encounter it for the
    second time, its second operation is placed on its corresponding machine,
    and so on.
    """

    def __init__(self, instance: Instance) -> None:
        """
        Instantiate the operation based encoding.

        :param instance: the JSSP instance
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        self.__machine_idx: Final[np.ndarray] = \
            np.zeros(instance.machines,
                     int_range_to_dtype(-1, instance.jobs - 1))
        self.__job_time: Final[np.ndarray] = \
            np.zeros(instance.jobs,
                     int_range_to_dtype(0, instance.makespan_upper_bound))
        self.__job_idx: Final[np.ndarray] = \
            np.zeros(instance.jobs,
                     int_range_to_dtype(0, instance.machines - 1))
        self.__instance: Final[Instance] = instance

    def decode(self, x: np.ndarray, y: np.ndarray) -> None:  # +book
        """
        Map an operation-based encoded array to a Gantt chart.

        :param x: the array
        :param y: the Gantt chart
        """
        decode(x, self.__machine_idx, self.__job_time,  # +book
               self.__job_idx, self.__instance, y)  # +book

    def __str__(self) -> str:
        """
        Get the name of this encoding.

        :return: `"operation_based_encoding"`
        :rtype: str
        """
        return "operation_based_encoding"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value(KEY_NUMPY_TYPE_MACHINE_IDX,
                         numpy_type_to_str(self.__machine_idx.dtype))
        logger.key_value(KEY_NUMPY_TYPE_JOB_IDX,
                         numpy_type_to_str(self.__job_idx.dtype))
        logger.key_value(KEY_NUMPY_TYPE_JOB_TIME,
                         numpy_type_to_str(self.__job_time.dtype))



"""
In this package, we provide implementations for basic search operators.

- In :mod:`~moptipy.operators.bitstrings`, you can find operators for elements
  from the :mod:`~moptipy.spaces.bitstrings` space.
- In :mod:`~moptipy.operators.ordered_choices`, you can find operators for
  elements from the :mod:`~moptipy.spaces.ordered_choices` space.
- In :mod:`~moptipy.operators.permutations`, you can find operators for
  elements from the :mod:`~moptipy.spaces.permutations` space.
- In :mod:`~moptipy.operators.vectors`, you can find operators for
  elements from the :mod:`~moptipy.spaces.vectorspace`.
"""



"""
The FFA-based version of the (1+1)-EA: the (1+1)-FEA.

This algorithm is based on :class:`~moptipy.algorithms.so.rls.RLS`, i.e., the
(1+1)-EA, but uses Frequency Fitness Assignment (FFA) as fitness assignment
process. FFA replaces all objective values with their encounter frequencies in
the selection decisions. The more often an objective value is encountered, the
higher gets its encounter frequency. Therefore, local optima are slowly
receiving worse and worse fitness.

Most of the existing metaheuristic algorithms have in common that they
maintain a set `Si` of one or multiple solutions and derive a set `Ni` of one
or multiple new solutions in each iteration `i`. From the joint set
`Pi = Si + Ni` of old and new solutions, they then select the set `Si+1` of
solutions to be propagated to the next iteration, and so on. This selection
decision is undertaken based mainly on the objective values `f(x)` of the
solutions `x in Pi` and solutions with better objective values tend to be
preferred over solutions with worse objective values.

Frequency Fitness Assignment (FFA) completely breaks with this most
fundamental concept of optimization. FFA was first proposed by
Weise *et al.* as a "plug-in" for metaheuristics intended to prevent
premature convergence. It therefore maintains a frequency table `H` for
objective values. Before the metaheuristic chooses the set `Si+1` from `Pi`,
it increments the encounter frequencies of the objective value of each
solution in `Pi`, i.e., performs `H[yj] <- H[yj] + 1` for each `xj in Pi`,
where `yj = f(xj)`. In its selection decisions, the algorithm then uses the
frequency fitness `H[yj]` instead of the objective values `yj`.

Here we integrate FFA into the randomized local search algorithm
:class:`~moptipy.algorithms.so.rls.RLS`, which is also known as the
`(1+1) EA`. In its original form, RLS maintains a single solution and derives
a slightly modified copy from it in every iteration. If the modified copy is
not worse than the original solution, it replaces it. "Not worse" here means
that its objective value needs to be better or equally good, i.e., `<=`, than
the maintained current best solution. The RLS with FFA (here called
`(1+1) FEA`) now replaces the comparison of objective values with a comparison
of the frequencies of the objective values. Of course, following the
definition of FFA, the frequencies are first incremented (both of the current
and the new solution) and then compared.

The algorithm is here implemented in two different ways: If the objective
function is always integer valued and the difference between its upper and
lower bound is not too high, then we count the frequency fitness by using a
numpy array. This means that frequency updates and getting frequency values is
very fast. If the objective function is not always integer or if the
difference between its maximum and minimum is too large, then we will use
a :class:`collections.Counter` to back the frequency table instead. This will
be slower and probably require more memory, but it may be the only way to
accommodate the frequency table. Of course, this will still fail if there are
too many different objective values and the memory consumed is simply too
high.

FFA is also implemented as a fitness assignment process
(:mod:`~moptipy.algorithms.so.fitness`) in module
:mod:`~moptipy.algorithms.so.fitnesses.ffa`.

1. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020.
   https://dx.doi.org/10.1109/TEVC.2020.3032090
2. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*. 2022.
   Early Access. https://dx.doi.org/10.1109/TEVC.2022.3191698
3. Thomas Weise, Mingxu Wan, Ke Tang, Pu Wang, Alexandre Devert, and Xin
   Yao. Frequency Fitness Assignment. *IEEE Transactions on Evolutionary
   Computation (IEEE-EC),* 18(2):226-243, April 2014.
   https://dx.doi.org/10.1109/TEVC.2013.2251885
4. Thomas Weise, Yan Chen, Xinlu Li, and Zhize Wu. Selecting a diverse set of
   benchmark instances from a tunable model problem for black-box discrete
   optimization algorithms. *Applied Soft Computing Journal (ASOC),*
   92:106269, June 2020. https://dx.doi.org/10.1016/j.asoc.2020.106269
5. Thomas Weise, Xinlu Li, Yan Chen, and Zhize Wu. Solving Job Shop Scheduling
   Problems Without Using a Bias for Good Solutions. In *Genetic and
   Evolutionary Computation Conference Companion (GECCO'21 Companion),*
   July 10-14, 2021, Lille, France. ACM, New York, NY, USA.
   ISBN 978-1-4503-8351-6. https://dx.doi.org/10.1145/3449726.3463124
6. Thomas Weise, Mingxu Wan, Ke Tang, and Xin Yao. Evolving Exact Integer
   Algorithms with Genetic Programming. In *Proceedings of the IEEE Congress
   on Evolutionary Computation (CEC'14), Proceedings of the 2014 World
   Congress on Computational Intelligence (WCCI'14),* pages 1816-1823,
   July 6-11, 2014, Beijing, China. Los Alamitos, CA, USA: IEEE Computer
   Society Press. ISBN: 978-1-4799-1488-3.
   https://dx.doi.org/10.1109/CEC.2014.6900292
"""
from collections import Counter
from io import StringIO
from typing import Callable, Final, Iterable, cast

import numpy as np
from numpy.random import Generator
from pycommons.io.csv import CSV_SEPARATOR
from pycommons.strings.string_conv import num_to_str
from pycommons.types import type_error

from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process
from moptipy.utils.nputils import DEFAULT_INT

#: the log section for the frequency table
H_LOG_SECTION: Final[str] = "H"

#: The difference between upper- and lower bound at which we switch from
#: using arrays as backing store for the frequency table H to maps.
SWITCH_TO_MAP_RANGE: Final[int] = 67_108_864


def _fea_flat(process: Process, op0: Callable, op1: Callable,
              lb: int, ub: int, log_h_tbl: bool) -> None:
    """
    Apply the (1+1)-FEA to an optimization problem.

    :param process: the black-box process object
    :param op0: the nullary search operator
    :param op1: the unary search operator
    :param lb: the lower bound
    :param ub: the upper bound
    :param log_h_tbl: should we log the H table?
    """
    # Create records for old and new point in the search space.
    best_x = process.create()  # record for best-so-far solution
    new_x = process.create()  # record for new solution

# h holds the encounter frequency of each objective value.
    h: Final[np.ndarray] = np.zeros(ub - lb + 1, DEFAULT_INT)
# Obtain the random number generator.
    random: Final[Generator] = process.get_random()

# Put function references in variables to save time.
    evaluate: Final[Callable] = process.evaluate  # the objective
    should_terminate: Final[Callable] = process.should_terminate

# Start at a random point in the search space and evaluate it.
    op0(random, best_x)  # Create 1 solution randomly and
    best_f: int = cast(int, evaluate(best_x)) - lb  # evaluate it.

    while not should_terminate():  # Until we need to quit...
        op1(random, new_x, best_x)  # new_x = neighbor of best_x
        new_f: int = cast(int, evaluate(new_x)) - lb

        h[new_f] += 1  # Increase the frequency of new_f and
        h[best_f] += 1  # of best_f.
        if h[new_f] <= h[best_f]:  # frequency of new_f no worse than best_f?
            best_f = new_f  # Store its objective value.
            best_x, new_x = new_x, best_x  # Swap best and new.

    if not log_h_tbl:
        return  # we are done here

    # After we are done, we want to print the H table.
    if h[best_f] == 0:  # Fix the H table for the case that only one
        h[best_f] = 1   # single FE was performed.
    log_h(process, range(len(h)), cast(Callable[[int | float], int],
                                       h.__getitem__),
          cast(Callable[[int | float], str],  # add the lower bound back in
               lambda i, _lb=lb: str(i + _lb)))


def _fea_map(process: Process, op0: Callable, op1: Callable,
             log_h_tbl: bool) -> None:
    """
    Apply the (1+1)-FEA to an optimization problem.

    :param process: the black-box process object
    :param op0: the nullary search operator
    :param op1: the unary search operator
    :param log_h_tbl: should we log the H table?
    """
    # Create records for old and new point in the search space.
    best_x = process.create()  # record for best-so-far solution
    new_x = process.create()  # record for new solution

# h holds the encounter frequency of each objective value.
    h: Final[Counter] = Counter()
# Obtain the random number generator.
    random: Final[Generator] = process.get_random()

# Put function references in variables to save time.
    evaluate: Final[Callable] = process.evaluate  # the objective
    should_terminate: Final[Callable] = process.should_terminate

# Start at a random point in the search space and evaluate it.
    op0(random, best_x)  # Create 1 solution randomly and
    best_f: int | float = evaluate(best_x)  # evaluate it.

    while not should_terminate():  # Until we need to quit...
        op1(random, new_x, best_x)  # new_x = neighbor of best_x
        new_f: int | float = evaluate(new_x)

        h[new_f] += 1  # Increase the frequency of new_f and
        h[best_f] += 1  # of best_f.
        if h[new_f] <= h[best_f]:  # frequency of new_f no worse than best_f?
            best_f = new_f  # Store its objective value.
            best_x, new_x = new_x, best_x  # Swap best and new.

    if not log_h_tbl:
        return

# After we are done, we want to print the H table.
    if h[best_f] == 0:  # Fix the H table for the case that only one
        h[best_f] = 1   # single FE was performed.
    log_h(process, h.keys(),
          cast(Callable[[int | float], int], h.__getitem__), num_to_str)


class FEA1plus1(Algorithm1):
    """
    The FFA-based version of the (1+1)-EA: the (1+1)-FEA.

    This algorithm applies Frequency Fitness Assignment (FFA).
    This means that it does not select solutions based on whether
    they are better or worse. Instead, it selects the solution whose
    objective value has been encountered during the search less often.
    The word "best" therefore is not used in the traditional sense, i.e.,
    that one solution is better than another one terms of its objective
    value. Instead, the current best solution is always the one whose
    objective value we have seen the least often.

    In each step, a (1+1)-FEA creates a modified copy `new_x` of the
    current best solution `best_x`. It then increments the frequency fitness
    of both solutions by 1. If frequency fitness of `new_x` is not bigger
    the one of `best_x`, it becomes the new `best_x`.
    Otherwise, it is discarded.

    This algorithm implementation requires that objective values are
    integers and have lower and upper bounds that are not too far
    away from each other. A more general version is available as a fitness
    assignment process (:mod:`~moptipy.algorithms.so.fitness`) that can
    be plugged into a general EA (:mod:`~moptipy.algorithms.so.general_ea`)
    in module :mod:`~moptipy.algorithms.so.fitnesses.ffa`.
    """

    def __init__(self, op0: Op0, op1: Op1, log_h_tbl: bool = True) -> None:
        """
        Create the (1+1)-FEA.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param log_h_tbl: should we log the H table?
        """
        super().__init__("fea1p1", op0, op1)
        if not isinstance(log_h_tbl, bool):
            raise type_error(log_h_tbl, "log_h_tbl", bool)
        #: True if we should log the H table, False otherwise
        self.log_h_tbl: Final[bool] = log_h_tbl

    def solve(self, process: Process) -> None:
        """
        Apply the (1+1)-FEA to an optimization problem.

        :param process: the black-box process object
        """
        if process.is_always_integer():
            lb: Final[int | float] = process.lower_bound()
            ub: Final[int | float] = process.upper_bound()
            if isinstance(ub, int) and isinstance(lb, int) \
                    and ((ub - lb) <= SWITCH_TO_MAP_RANGE):
                _fea_flat(process, self.op0.op0, self.op1.op1, lb, ub,
                          self.log_h_tbl)
                return
        _fea_map(process, self.op0.op0, self.op1.op1, self.log_h_tbl)


def __h_to_str(indices: Iterable[int | float],
               h: Callable[[int | float], int],
               print_index: Callable[[int | float], str]) -> str:
    """
    Convert a frequency table `H` to a string.

    :param indices: the iterable of indices
    :param h: the history table
    :param print_index: a function to print an index
    :returns: a string representation of the `H` table

    >>> hl = [0, 0, 1, 7, 4, 0, 0, 9, 0]
    >>> __h_to_str(range(len(hl)), hl.__getitem__, str)
    '2;1;3;7;4;4;7;9'
    >>> __h_to_str(range(len(hl)), hl.__getitem__, lambda ii: str(ii + 1))
    '3;1;4;7;5;4;8;9'
    >>> hd = {1: 5, 4: 7, 3: 6, 2: 9}
    >>> __h_to_str(sorted(hd.keys()), hd.__getitem__, str)
    '1;5;2;9;3;6;4;7'
    >>> hd = Counter({1: 5, 4: 7, 3: 6, 2: 9})
    >>> __h_to_str(sorted(hd.keys()), hd.__getitem__, str)
    '1;5;2;9;3;6;4;7'
    >>> try:
    ...     hd = {1: 0}
    ...     __h_to_str(sorted(hd.keys()), hd.__getitem__, str)
    ... except ValueError as ve:
    ...     print(ve)
    empty H table?
    >>> hx = np.zeros(100, int)
    >>> hx[10] = 4
    >>> hx[12] = 234
    >>> hx[89] = 111
    >>> hx[45] = 2314
    >>> __h_to_str(range(100), hx.__getitem__, str)
    '10;4;12;234;45;2314;89;111'
    >>> __h_to_str(range(100), hx.__getitem__, lambda k: str(10 + k))
    '20;4;22;234;55;2314;99;111'
    >>> hx = np.zeros(100, np.int8)
    >>> hx[10] = 4
    >>> hx[12] = 34
    >>> hx[89] = 11
    >>> hx[45] = 14
    >>> __h_to_str(range(100), hx.__getitem__, str)
    '10;4;12;34;45;14;89;11'
    >>> __h_to_str(range(100), hx.__getitem__, lambda k: str(10 + k))
    '20;4;22;34;55;14;99;11'
    >>> hx = np.zeros(100, np.uint64)
    >>> hx[10] = 4232124356792834738
    >>> hx[12] = 3423443534534
    >>> hx[89] = 13589732857375734566
    >>> hx[45] = 14
    >>> __h_to_str(range(100), hx.__getitem__, str)
    '10;4232124356792834738;12;3423443534534;45;14;89;13589732857375734566'
    >>> __h_to_str(range(100), hx.__getitem__, lambda k: str(10 + k))
    '20;4232124356792834738;22;3423443534534;55;14;99;13589732857375734566'
    """
    with StringIO() as out:
        write: Callable[[str], int] = out.write  # fast call
        csep: Final[str] = CSV_SEPARATOR
        sep: str = ""
        for i in indices:
            v = h(i)
            if v > 0:
                write(sep)
                sep = csep
                write(print_index(i))
                write(sep)
                write(str(v))
        res: Final[str] = out.getvalue()
    if len(res) <= 0:
        raise ValueError("empty H table?")
    return res


def log_h(process: Process, indices: Iterable[int | float],
          h: Callable[[int | float], int],
          print_index: Callable[[int | float], str]) -> None:
    """
    Convert a frequency table `H` to a string and log it to a process.

    The frequency table is logged as a single line of text into a section
    `H` delimited by the lines `BEGIN_H` and `END_H`. The line consists
    of `2*n` semicolon separated values. Each such value pair consists of
    an objective value `y` and its observed frequency `H[y]`. The former is
    either an integer or a float and the latter is an integer.

    :param process: the process
    :param indices: the iterable of indices
    :param h: the history table
    :param print_index: a function to print an index
    """
    if process.has_log():
        s = __h_to_str(indices, h, print_index)
        if len(s) > 0:
            process.add_log_section(H_LOG_SECTION, s)



"""
Search operators for signed permutations with repetitions.

These operators can be used in conjunction with the space
:mod:`~moptipy.spaces.signed_permutations`.

- Module :mod:`~moptipy.operators.signed_permutations.op0_shuffle_and_flip`
  provides a nullary operator sampling an entirely random signed permutation.
- Module :mod:`~moptipy.operators.signed_permutations.op1_swap_2_or_flip`
  provides a unary operator which either swaps exactly two different elements
  of a permutation *or* flips the sign of one element.
"""
